from collections.abc import Iterable, Sequence
from typing import Literal

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    RequestMessage,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.request_classes.filters import StreamlitFilter
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.client.resource_classes.streamlit_ import (
    StreamlitRequest,
    StreamlitResponse,
)


class StreamlitAPI(CDFResourceAPI[StreamlitResponse]):
    """API for managing Streamlit apps in CDF.

    Streamlit apps are stored as file metadata objects with a specific directory prefix
    and app properties encoded in the metadata field.
    """

    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(method="POST", path="/files", item_limit=1, concurrency_max_workers=1),
                "retrieve": Endpoint(method="POST", path="/files/byids", item_limit=1000, concurrency_max_workers=1),
                "update": Endpoint(method="POST", path="/files/update", item_limit=1000, concurrency_max_workers=1),
                "delete": Endpoint(method="POST", path="/files/delete", item_limit=1000, concurrency_max_workers=1),
                "list": Endpoint(method="POST", path="/files/list", item_limit=1000),
            },
        )

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[StreamlitResponse]:
        return PagedResponse[StreamlitResponse].model_validate_json(response.body)

    def _reference_response(self, response: SuccessResponse) -> ResponseItems[ExternalId]:
        return ResponseItems[ExternalId].model_validate_json(response.body)

    def create(self, items: Sequence[StreamlitRequest], overwrite: bool = False) -> list[StreamlitResponse]:
        """Create Streamlit apps in CDF.

        Args:
            items: List of StreamlitRequest objects to create.
            overwrite: Whether to overwrite existing apps with the same external ID.

        Returns:
            List of created StreamlitResponse objects.
        """
        # The Streamlit API is a wrapper of the File API, which is different from other APIs,
        # thus we have a custom implementation here.
        # - It only allow one item per request that is not wrapped in an "items" field.
        # - It uses a query parameter for "overwrite" instead of including it in the body
        endpoint = self._method_endpoint_map["create"]
        results: list[StreamlitResponse] = []
        for item in items:
            request = RequestMessage(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                body_content=item.dump(),
                parameters={"overwrite": overwrite},
            )
            response = self._http_client.request_single_retries(request)
            result = response.get_success_or_raise()
            results.append(StreamlitResponse.model_validate_json(result.body))
        return results

    def retrieve(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> list[StreamlitResponse]:
        """Retrieve Streamlit apps from CDF.

        Args:
            items: List of ExternalId objects to retrieve.
            ignore_unknown_ids: Whether to ignore unknown IDs.

        Returns:
            List of retrieved StreamlitResponse objects.
        """
        return self._request_item_response(
            items, method="retrieve", extra_body={"ignoreUnknownIds": ignore_unknown_ids}
        )

    def update(
        self, items: Sequence[StreamlitRequest], mode: Literal["patch", "replace"] = "replace"
    ) -> list[StreamlitResponse]:
        """Update Streamlit apps in CDF.

        Args:
            items: List of StreamlitRequest objects to update.
            mode: Update mode, either "patch" or "replace".

        Returns:
            List of updated StreamlitResponse objects.
        """
        return self._update(items, mode=mode)

    def delete(self, items: Sequence[ExternalId], ignore_unknown_ids: bool = False) -> None:
        """Delete Streamlit apps from CDF.

        Args:
            items: List of ExternalId objects to delete.
            ignore_unknown_ids: Whether to ignore unknown IDs.
        """
        self._request_no_response(items, "delete", extra_body={"ignoreUnknownIds": ignore_unknown_ids})

    def paginate(
        self,
        filter: StreamlitFilter | None = None,
        limit: int = 100,
        cursor: str | None = None,
    ) -> PagedResponse[StreamlitResponse]:
        """Paginate over Streamlit apps in CDF.

        Args:
            filter: Filter by data set IDs and/or asset subtree IDs.
            limit: Maximum number of items to return per page.
            cursor: Cursor for pagination.

        Returns:
            PagedResponse of StreamlitResponse objects.
        """
        return self._paginate(
            cursor=cursor,
            limit=limit,
            body={"filter": (filter or StreamlitFilter()).dump()},
        )

    def iterate(
        self,
        filter: StreamlitFilter | None = None,
        limit: int | None = 100,
    ) -> Iterable[list[StreamlitResponse]]:
        """Iterate over all Streamlit apps in CDF.

        Args:
            filter: Filter by data set IDs and/or asset subtree IDs.
            limit: Maximum number of items to return per page.

        Returns:
            Iterable of lists of StreamlitResponse objects.
        """
        return self._iterate(
            limit=limit,
            body={"filter": (filter or StreamlitFilter()).dump()},
        )

    def list(
        self,
        filter: StreamlitFilter | None = None,
        limit: int | None = 100,
    ) -> list[StreamlitResponse]:
        """List all Streamlit apps in CDF.

        Args:
            filter: StreamlitFilter to filter the apps.
            limit: Maximum number of items to return. None for all items.

        Returns:
            List of StreamlitResponse objects.
        """
        return self._list(
            limit=limit,
            body={"filter": (filter or StreamlitFilter()).dump()},
        )



from typing import Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId
from cognite_toolkit._cdf_tk.constants import INSTANCE_EXTERNAL_ID_PATTERN

from .base import ToolkitResource
from .view_field_definitions import ViewReference


class ResourceViewMappingYAML(ToolkitResource):
    external_id: str = Field(
        description="External-id of the mapping.",
        min_length=1,
        max_length=256,
        pattern=INSTANCE_EXTERNAL_ID_PATTERN,
    )
    resource_type: Literal["asset", "event", "file", "timeseries", "sequence", "assetAnnotation", "fileAnnotation"] = (
        Field(
            description="The type of the resource to map to the view.",
        )
    )
    view_id: ViewReference = Field(
        description="The view to map the resource to.",
    )
    property_mapping: dict[str, str] = Field(
        description="A dictionary mapping from resource property to view property.",
    )

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)



"""Principal resource classes for the Cognite Principals API.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Principals
"""

from abc import ABC
from typing import Annotated, Literal, TypeAlias

from pydantic import Field

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import PrincipalId

PrincipalType: TypeAlias = Literal["SERVICE_ACCOUNT", "USER"]


class CreatedBy(BaseModelObject):
    org_id: str
    user_id: str


class PrincipalDefinition(BaseModelObject, ABC):
    id: str
    type: Literal["SERVICE_ACCOUNT", "USER"]
    name: str
    picture_url: str

    def as_id(self) -> PrincipalId:
        """Convert the principal definition to a PrincipalId."""
        return PrincipalId(id=self.id)


class ServiceAccountPrincipal(PrincipalDefinition):
    type: Literal["SERVICE_ACCOUNT"] = "SERVICE_ACCOUNT"
    description: str | None = None
    external_id: str | None = None
    created_by: CreatedBy
    created_time: int
    last_updated_time: int


class UserPrincipal(PrincipalDefinition):
    type: Literal["USER"] = "USER"
    email: str | None = None
    given_name: str | None = None
    middle_name: str | None = None
    family_name: str | None = None


Principal = Annotated[
    ServiceAccountPrincipal | UserPrincipal,
    Field(discriminator="type"),
]


class LoginSession(BaseModelObject):
    id: str
    created_time: int
    status: Literal["ACTIVE", "LOGGED_OUT", "EXPIRED", "REVOKED"]
    deactivated_time: int | None = None
    # This is not part of the API response, but we include it here to track which principal the session belongs to.
    principal: str = Field("")



from collections.abc import Callable

from cognite_toolkit._cdf_tk.commands.build_v2.data_classes._module import Module, SuccessfulReadResource
from cognite_toolkit._cdf_tk.rules._base import ToolkitResourceRule
from cognite_toolkit._cdf_tk.utils._auxiliary import get_concrete_subclasses

_rules_registry: dict[str, list[type[ToolkitResourceRule]]] | None = None


def get_rules_registry(force_reload: bool = False) -> dict[str, list[type[ToolkitResourceRule]]]:
    """Get the registry of rules, optionally forcing a reload.

    Args:
        force_reload: If True, forces a reload of the rules registry.

    Returns:
        A dictionary mapping resource types to lists of ToolkitRule classes.
    """
    global _rules_registry
    if _rules_registry is None or force_reload:
        registry: dict[str, list[type[ToolkitResourceRule]]] = {}
        rules = get_concrete_subclasses(ToolkitResourceRule)  # type: ignore

        for rule_cls in rules:
            registry.setdefault(rule_cls.resource_type, []).append(rule_cls)
        _rules_registry = registry
    return _rules_registry


class RulesOrchestrator:
    def __init__(
        self,
        can_run_validator: Callable[[str, type], bool] | None = None,
        enable_alpha_validators: bool = False,
    ) -> None:
        self._can_run_validator = can_run_validator or (lambda code, issue_type: True)
        self._enable_alpha_validators = enable_alpha_validators

    def run(self, module: Module) -> None:
        """Run all applicable rules on the provided modules while updating modules insights.

        Args:
            module: The module to run the rules on.
        """

        rules_registry = get_rules_registry()

        for resource in module.resources:
            if not isinstance(resource, SuccessfulReadResource):
                continue
            for rule in rules_registry.get(resource.resource_type.kind, []):
                if rule.alpha and not self._enable_alpha_validators:
                    continue

                if self._can_run_validator(rule.code, rule.insight_type):
                    if insights := rule([resource.resource]).validate():
                        module.insights.extend(insights)



from typing import ClassVar, Literal

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, ResponseResource, UpdatableRequestResource
from cognite_toolkit._cdf_tk.client._types import Metadata

from .identifiers import ExternalId


class Asset(BaseModelObject):
    external_id: str | None = None
    name: str
    parent_id: int | None = None
    parent_external_id: str | None = None
    description: str | None = None
    metadata: Metadata | None = None
    data_set_id: int | None = None
    source: str | None = None
    labels: list[dict[Literal["externalId"], str]] | None = None
    geo_location: dict[str, JsonValue] | None = None

    def as_id(self) -> ExternalId:
        if self.external_id is None:
            raise ValueError("Cannot convert AssetRequest to ExternalId when external_id is None")
        return ExternalId(external_id=self.external_id)


class AssetRequest(Asset, UpdatableRequestResource):
    container_fields: ClassVar[frozenset[str]] = frozenset({"metadata", "labels"})
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset({"parent_id", "parent_external_id"})


class AssetAggregateItem(BaseModelObject):
    child_count: int
    depth: int
    path: list[dict[Literal["id"], int]]


class AssetResponse(Asset, ResponseResource[AssetRequest]):
    created_time: int
    last_updated_time: int
    root_id: int
    aggregates: AssetAggregateItem | None = None
    id: int

    @classmethod
    def request_cls(cls) -> type[AssetRequest]:
        return AssetRequest



"""Base classes for Cognite Client resources and identifiers."""

import sys
import types
from abc import ABC, abstractmethod
from functools import total_ordering
from typing import Annotated, Any, ClassVar, Generic, Literal, TypeVar, Union, get_args, get_origin

from pydantic import BaseModel, ConfigDict
from pydantic.alias_generators import to_camel
from pydantic_core import PydanticUndefined

from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.utils.file import read_yaml_content, yaml_safe_dump

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class BaseModelObject(BaseModel):
    """Base class for all object. This includes resources and nested objects."""

    # We allow extra fields to support forward compatibility.
    model_config = ConfigDict(
        alias_generator=to_camel, extra="allow" if Flags.v08.is_enabled() else "ignore", populate_by_name=True
    )

    def model_post_init(self, __context: Any) -> None:
        super().model_post_init(__context)
        for field_name, field_info in type(self).model_fields.items():
            if field_name in self.model_fields_set:
                continue
            if field_info.default is PydanticUndefined:
                continue
            if not _is_optional(field_info.annotation):
                self.__pydantic_fields_set__.add(field_name)

    def dump(self, camel_case: bool = True, exclude_extra: bool = False) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.
            exclude_extra (bool): Whether to exclude extra fields not defined in the model. Default is False.

        """
        if exclude_extra:
            return self.model_dump(
                mode="json",
                by_alias=camel_case,
                exclude_unset=True,
                exclude=set(self.__pydantic_extra__) if self.__pydantic_extra__ else None,
            )
        return self.model_dump(mode="json", by_alias=camel_case, exclude_unset=True)

    def dump_yaml(self, camel_case: bool = True, exclude_extra: bool = False) -> str:
        """Dump the resource to a YAML string.

        Args:
            camel_case (bool): Whether to use camelCase for the keys. Default is True.
            exclude_extra (bool): Whether to exclude extra fields not defined in the model. Default is False.

        """
        return yaml_safe_dump(self.dump(camel_case=camel_case, exclude_extra=exclude_extra))

    @classmethod
    def _load(cls, resource: dict[str, Any]) -> Self:
        """Load method to match CogniteResource signature."""
        return cls.model_validate(resource, by_alias=True)

    @classmethod
    def load_yaml(cls, yaml_content: str) -> Self:
        """Load the resource from a YAML string."""
        content = read_yaml_content(yaml_content)
        if not isinstance(content, dict):
            raise ValueError(f"YAML content must be a dictionary to load into {cls}, got {type(content)}")
        return cls._load(content)


T_BaseModelObject = TypeVar("T_BaseModelObject", bound=BaseModelObject)


class RequestItem(BaseModelObject, ABC):
    """A request item is any object that can be sent to the CDF API as part of a request."""

    def __str__(self) -> str:
        """All request items must implement a string representation.

        This is used to identify the item in error messages and logs.
        """
        raise NotImplementedError()


T_RequestItem = TypeVar("T_RequestItem", bound=RequestItem)


@total_ordering
class Identifier(RequestItem):
    """Base class for all identifier objects typically
    {"externalId": "..."}, {"id": ...}, {"space": "...", "externalId: "..."}."""

    model_config = ConfigDict(alias_generator=to_camel, extra="ignore", frozen=True)

    def dump(self, camel_case: bool = True, exclude_extra: bool = False) -> dict[str, Any]:
        """Dump the resource to a dictionary.

        This is the default serialization method for request resources.
        """
        return self.model_dump(mode="json", by_alias=camel_case, exclude_unset=True)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Identifier):
            return NotImplemented
        return self.dump() == other.dump()

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Identifier):
            return NotImplemented
        return str(self) < str(other)


T_Identifier = TypeVar("T_Identifier", bound=Identifier)


class RequestResource(RequestItem, ABC):
    @abstractmethod
    def as_id(self) -> Identifier:
        raise NotImplementedError()

    def __str__(self) -> str:
        return str(self.as_id())


T_RequestResource = TypeVar("T_RequestResource", bound=RequestResource)


class UpdatableRequestResource(RequestResource, ABC):
    container_fields: ClassVar[frozenset[str]] = frozenset()
    non_nullable_fields: ClassVar[frozenset[str]] = frozenset()

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        """Convert the request resource to an update item."""
        update_item = self.as_id().dump(camel_case=True)
        update: dict[str, Any] = {}
        field_by_name = {info.alias or field_id: (field_id, info) for field_id, info in type(self).model_fields.items()}
        # When mode is "patch", we only include fields that are set
        exclude_unset = mode == "patch"
        for key, value in self.model_dump(mode="json", by_alias=True, exclude_unset=exclude_unset).items():
            if key in update_item:
                # Skip identifier fields
                continue
            if key not in field_by_name:
                # Skip unknown fields
                continue
            field_id, info = field_by_name[key]
            if field_id in self.container_fields:
                if mode == "patch":
                    update[key] = {"add": value}
                elif mode == "replace":
                    if value is None:
                        origin = _get_annotation_origin(info.annotation)
                        if origin is list:
                            update[key] = {"set": []}
                        elif origin is dict:
                            update[key] = {"set": {}}
                        else:
                            raise NotImplementedError(
                                f'Cannot replace container field "{key}" with None when its type is unknown.'
                            )
                    else:
                        update[key] = {"set": value}
                else:
                    raise NotImplementedError(f'Update mode "{mode}" is not supported for container fields.')
            elif value is None:
                if field_id not in self.non_nullable_fields:
                    update[key] = {"setNull": True}
            else:
                update[key] = {"set": value}
        update_item["update"] = update
        return update_item


def _is_optional(annotation: Any) -> bool:
    """Check if a type annotation includes None as a valid type."""
    origin = get_origin(annotation)
    # Check for Union type (both typing.Union and | syntax from Python 3.10+)
    is_union = origin is Union or isinstance(annotation, getattr(types, "UnionType", ()))
    if is_union:
        return type(None) in get_args(annotation)
    return False


def _get_annotation_origin(field_type: Any) -> Any:
    origin = get_origin(field_type)
    args = get_args(field_type)

    # Handle Annotated types by extracting the underlying type
    if origin is Annotated and args:
        field_type = args[0]
        origin = get_origin(field_type)
        args = get_args(field_type)

    # Check for Union type (both typing.Union and | syntax from Python 3.10+)
    is_union = origin is Union or isinstance(field_type, getattr(types, "UnionType", ()))

    if is_union:
        # Handle Optional[T] by filtering out NoneType
        none_types = (type(None), types.NoneType)
        non_none_args = [arg for arg in args if arg not in none_types]
        if len(non_none_args) == 1:
            field_type = non_none_args[0]
            origin = get_origin(field_type) or field_type

            if origin is Annotated:
                return _get_annotation_origin(field_type)

    return origin


class ResponseResource(BaseModelObject, Generic[T_RequestResource], ABC):
    @classmethod
    @abstractmethod
    def request_cls(cls) -> type[T_RequestResource]:
        """Return the class of the corresponding request resource."""
        raise NotImplementedError()

    def as_request_resource(self) -> T_RequestResource:
        """Convert the response resource to a request resource."""
        request_cls = self.request_cls()
        return request_cls.model_validate(self.dump(), extra="ignore", by_alias=True)

    # Todo remove when CogniteClient data classes are completely removed from the codebase
    # and we only use the pydantic resource classes instead.from
    def as_write(self) -> T_RequestResource:
        """Alias for as_request_resource to match protocol signature."""
        return self.as_request_resource()


T_ResponseResource = TypeVar("T_ResponseResource", bound=ResponseResource)



from abc import ABC
from typing import Any, Literal

from pydantic import Field, JsonValue, field_serializer, model_validator
from pydantic_core.core_schema import FieldSerializationInfo

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedViewReference

from ._data_types import DirectNodeRelation
from ._references import ContainerReference, ViewReference
from ._view_property import (
    MultiReverseDirectRelationPropertyResponse,
    SingleReverseDirectRelationPropertyResponse,
    ViewCorePropertyRequest,
    ViewCorePropertyResponse,
    ViewRequestProperty,
    ViewResponseProperty,
)


class View(BaseModelObject, ABC):
    space: str
    external_id: str
    version: str
    name: str | None = None
    description: str | None = None
    filter: JsonValue | None = None
    implements: list[ViewReference] | None = None

    def as_id(self) -> ViewReference:
        return ViewReference(space=self.space, external_id=self.external_id, version=self.version)

    def as_typed_id(self) -> TypedViewReference:
        return TypedViewReference(space=self.space, external_id=self.external_id, version=self.version)

    @model_validator(mode="before")
    def set_connection_type_on_primary_properties(cls, data: dict[str, Any]) -> dict[str, Any]:
        if "properties" not in data:
            return data
        properties = data["properties"]
        if not isinstance(properties, dict):
            return data
        # We assume all properties without connectionType are core properties.
        # The reason we set connectionType is to make it easy for pydantic to discriminate the union.
        # This also leads to better error messages, as if there is a union and pydantic does not know which
        # type to pick it will give errors from all type in the union.
        new_properties: dict[str, Any] = {}
        for prop_id, prop in properties.items():
            if isinstance(prop, dict) and "connectionType" not in prop:
                prop_copy = prop.copy()
                prop_copy["connectionType"] = "primary_property"
                new_properties[prop_id] = prop_copy
            else:
                new_properties[prop_id] = prop
        if new_properties:
            new_data = data.copy()
            new_data["properties"] = new_properties
            return new_data

        return data

    @field_serializer("implements", mode="plain")
    @classmethod
    def serialize_implements(
        cls, implements: list[ViewReference] | None, info: FieldSerializationInfo
    ) -> list[dict[str, Any]] | None:
        if implements is None:
            return None
        output: list[dict[str, Any]] = []
        for view in implements:
            dumped = view.model_dump(**vars(info))
            dumped["type"] = "view"
            output.append(dumped)
        return output


class ViewRequest(View, RequestResource):
    properties: dict[str, ViewRequestProperty] | None = Field(
        default=None,
        description="View with included properties and expected edges, indexed by a unique space-local identifier.",
    )

    @property
    def used_containers(self) -> set[ContainerReference]:
        """Get all containers referenced by this view."""
        return {
            prop.container for prop in (self.properties or {}).values() if isinstance(prop, ViewCorePropertyRequest)
        }


class ViewResponse(View, ResponseResource[ViewRequest]):
    properties: dict[str, ViewResponseProperty]

    created_time: int
    last_updated_time: int
    writable: bool
    queryable: bool
    used_for: Literal["node", "edge", "all"]
    is_global: bool
    mapped_containers: list[ContainerReference]

    @classmethod
    def request_cls(cls) -> type[ViewRequest]:
        return ViewRequest

    def as_request_resource(self) -> ViewRequest:
        dumped = self.model_dump(by_alias=True, exclude={"properties"}, exclude_unset=True)
        properties: dict[str, Any] = {}
        for key, value in self.properties.items():
            if isinstance(value, ViewCorePropertyResponse) and isinstance(value.type, DirectNodeRelation):
                # Special case. In the request the source of DirectNodeRelation is set on the Property object,
                # while in the response it is set on the DirectNodeRelation object.
                request_object = value.as_request().model_dump(by_alias=True)
                request_object["source"] = value.type.source.model_dump(by_alias=True) if value.type.source else None
                properties[key] = request_object
            elif isinstance(
                value,
                ViewCorePropertyResponse
                | SingleReverseDirectRelationPropertyResponse
                | MultiReverseDirectRelationPropertyResponse,
            ):
                properties[key] = value.as_request().model_dump(by_alias=True)
            else:
                properties[key] = value.model_dump(by_alias=True)

        dumped["properties"] = properties
        return ViewRequest.model_validate(dumped, extra="ignore")

    @field_serializer("mapped_containers", mode="plain")
    @classmethod
    def serialize_mapped_containers(
        cls, mapped_containers: list[ContainerReference], info: FieldSerializationInfo
    ) -> list[dict[str, Any]]:
        return [container.model_dump(**vars(info)) | {"type": "container"} for container in mapped_containers]

    @field_serializer("properties", mode="plain")
    @classmethod
    def serialize_properties_special_handling_direct_relation_with_source(
        cls, properties: dict[str, ViewResponseProperty], info: FieldSerializationInfo
    ) -> dict[str, dict[str, Any]]:
        output: dict[str, dict[str, Any]] = {}
        for prop_id, prop in properties.items():
            output[prop_id] = prop.model_dump(**vars(info))
            if (
                isinstance(prop, ViewCorePropertyResponse)
                and isinstance(prop.type, DirectNodeRelation)
                and prop.type.source
            ):
                # We manually include the source as this is excluded by default. The reason why this is excluded
                # is that the DirectNodeRelation is used for both request and response, and in the request the source
                # does not exist on the DirectNodeRelation, but on the Property object.
                output[prop_id]["type"]["source"] = prop.type.source.model_dump(**vars(info)) | {"type": "view"}
        return output



from pathlib import Path

from cognite.client.utils._text import to_camel_case
from pydantic import BaseModel, model_validator
from rich import print

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.storageio._data_classes import ModelList

from ._base import ToolkitCommand


class RespaceMapping(BaseModel, alias_generator=to_camel_case, populate_by_name=True):
    """A single node respace mapping from source to target.

    The external ID is preserved — only the space changes.

    CSV columns: sourceSpace, externalId, targetSpace
    """

    source_space: str
    external_id: str
    target_space: str

    @model_validator(mode="after")
    def _target_differs_from_source(self) -> "RespaceMapping":
        if self.source_space == self.target_space:
            raise ValueError(
                f"targetSpace must differ from sourceSpace, "
                f"but both are '{self.source_space}' for externalId '{self.external_id}'"
            )
        return self


class RespaceMappingList(ModelList[RespaceMapping]):
    @classmethod
    def _get_base_model_cls(cls) -> type[RespaceMapping]:
        return RespaceMapping


class RespaceCommand(ToolkitCommand):
    def __init__(
        self,
        print_warning: bool = True,
        skip_tracking: bool = False,
        silent: bool = False,
        client: ToolkitClient | None = None,
    ):
        super().__init__(print_warning, skip_tracking, silent, client)

    def plan(self, csv_file: Path, output_file: Path) -> None:
        """Generate a respace plan from a CSV file."""
        print(f"[bold]Planning respace from:[/] {csv_file}")
        print(f"[bold]Output plan file:[/] {output_file}")
        print("[bold yellow]:construction: Work in Progress, you'll be able to plan soon! :construction:[/]")

    def execute(self, plan_file: Path, backup_dir: Path, dry_run: bool = False) -> None:
        """Execute a respace plan."""
        verb = "Would execute" if dry_run else "Executing"
        print(f"[bold]{verb} plan:[/] {plan_file}")
        print(f"[bold]Backup directory:[/] {backup_dir}")
        print("[bold yellow]:construction: Work in Progress, you'll be able to execute soon! :construction:[/]")



from typing import ClassVar, Literal

from pydantic import JsonValue

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId

from ._common import MapType


class RobotMap(BaseModelObject):
    """The map resource allows defining both context maps and robot maps of a specific location. A context map is a
    visual representation of a location, for example, a 3D model, a 2D floor plan, or a point cloud model.
    A robot map is a representation of where a robot is able to navigate. Maps need to be aligned with respect
    to each other using coordinate frames.

    Args:
        name: Map name.
        external_id: Map external id. Must be unique for the resource type.
        map_type: Map type.
        description: Description of Map. Textual description of the Map.
        frame_external_id: External id of the map's reference frame.
        data: Map-specific data.
        location_external_id: External id of the location.
        scale: Uniform scaling factor, for example, for map unit conversion (centimeter to meter).
    """

    external_id: str
    name: str
    map_type: MapType
    description: str | None = None
    frame_external_id: str | None = None
    data: dict[str, JsonValue] | None = None
    location_external_id: str | None = None
    scale: float | None = None

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class RobotMapRequest(RobotMap, UpdatableRequestResource):
    """Request resource for creating or updating a RobotMap."""

    non_nullable_fields: ClassVar[frozenset[str]] = frozenset(
        {"description", "data", "frame_external_id", "location_external_id", "scale"}
    )

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, JsonValue]:
        update = super().as_update(mode)
        update["update"].pop("mapType", None)
        return update


class RobotMapResponse(RobotMap, ResponseResource[RobotMapRequest]):
    """Response resource for a RobotMap."""

    created_time: int
    updated_time: int

    @classmethod
    def request_cls(cls) -> type[RobotMapRequest]:
        return RobotMapRequest



from cognite_toolkit._cdf_tk.client.api.instances import WrappedInstancesAPI
from cognite_toolkit._cdf_tk.client.cdf_client import PagedResponse, ResponseItems
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    ItemsSuccessResponse,
    SuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.cognite_file import (
    CogniteFileRequest,
    CogniteFileResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.instance_api import TypedNodeIdentifier


class CogniteFilesAPI(WrappedInstancesAPI[TypedNodeIdentifier, CogniteFileResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(http_client, CogniteFileRequest.VIEW_ID)

    def _validate_response(self, response: SuccessResponse) -> ResponseItems[TypedNodeIdentifier]:
        return ResponseItems[TypedNodeIdentifier].model_validate_json(response.body)

    def _validate_page_response(
        self, response: SuccessResponse | ItemsSuccessResponse
    ) -> PagedResponse[CogniteFileResponse]:
        return PagedResponse[CogniteFileResponse].model_validate_json(response.body)

    def list(self, spaces: list[str] | None = None, limit: int | None = 100) -> list[CogniteFileResponse]:
        """List all CogniteFile instances.

        Args:
            spaces: Optional list of spaces to filter by.
            limit: Maximum number of items to return. If None, all items are returned.

        Returns:
            List of CogniteFileResponse objects.
        """
        return super()._list_instances(spaces=spaces, instance_type="node", limit=limit)



from pathlib import Path
from typing import Annotated

from pydantic import Field, TypeAdapter, ValidationError

from cognite_toolkit._cdf_tk.feature_flags import Flags
from cognite_toolkit._cdf_tk.tk_warnings import MediumSeverityWarning, ToolkitWarning
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.utils.file import read_yaml_file
from cognite_toolkit._cdf_tk.validation import humanize_validation_error

from ._asset_centric import AssetCentricFileSelector, AssetCentricSelector, AssetSubtreeSelector, DataSetSelector
from ._base import DataSelector
from ._canvas import CanvasExternalIdSelector, CanvasSelector
from ._charts import AllChartsSelector, ChartExternalIdSelector, ChartOwnerSelector, ChartSelector
from ._datapoints import (
    DataPointsDataSetSelector,
    DataPointsFileSelector,
    DataPointsSelector,
    ExternalIdColumn,
    InstanceColumn,
    InternalIdColumn,
    TimeSeriesColumn,
)
from ._file_content import (
    FileContentSelector,
    FileDataModelingTemplate,
    FileDataModelingTemplateSelector,
    FileIdentifierSelector,
    FileMetadataTemplate,
    FileMetadataTemplateSelector,
)
from ._instances import (
    InstanceFileSelector,
    InstanceSelector,
    InstanceSpaceSelector,
    InstanceViewSelector,
    SelectedView,
)
from ._raw import RawTableSelector, SelectedTable
from ._records import RecordContainerSelector, SelectedContainer, SelectedStream
from ._three_d import ThreeDModelFilteredSelector, ThreeDModelIdSelector, ThreeDSelector

Selector = Annotated[
    RawTableSelector
    | ThreeDModelIdSelector
    | ThreeDModelFilteredSelector
    | InstanceViewSelector
    | InstanceFileSelector
    | InstanceSpaceSelector
    | AllChartsSelector
    | ChartOwnerSelector
    | AssetSubtreeSelector
    | AssetCentricFileSelector
    | DataSetSelector
    | DataPointsDataSetSelector
    | DataPointsFileSelector
    | ChartExternalIdSelector
    | CanvasExternalIdSelector
    | FileMetadataTemplateSelector
    | FileDataModelingTemplateSelector
    | FileIdentifierSelector
    | RecordContainerSelector,
    Field(discriminator="type"),
]

ALPHA_SELECTORS = {FileIdentifierSelector, RecordContainerSelector}
INTERNAL = {ThreeDModelIdSelector, ThreeDModelFilteredSelector}
SelectorAdapter: TypeAdapter[Selector] = TypeAdapter(Selector)


def load_selector(manifest_file: Path) -> Selector | ToolkitWarning:
    """Loads a selector from a manifest file.

    Args:
        manifest_file: Path to the manifest file.

    Returns:
        A selector object or a toolkit warning if loading fails or the selector is an alpha feature that is not enabled.
    """
    selector_dict = read_yaml_file(manifest_file, expected_output="dict")
    try:
        selector = SelectorAdapter.validate_python(selector_dict)
    except ValidationError as e:
        errors = humanize_validation_error(e)
        return ResourceFormatWarning(manifest_file, tuple(errors), text="Invalid selector in metadata file, skipping.")
    if not Flags.EXTEND_UPLOAD.is_enabled() and type(selector) in ALPHA_SELECTORS:
        return MediumSeverityWarning(
            f"Selector type '{type(selector).__name__}' in file '{manifest_file}' is in alpha. To enable it set the alpha flag 'extend-upload = true' in your CDF.toml file."
        )
    elif type(selector) in INTERNAL:
        return MediumSeverityWarning(
            f"Selector type '{type(selector).__name__}' in file '{manifest_file}' is for internal use only and cannot be used."
        )
    return selector


__all__ = [
    "AllChartsSelector",
    "AssetCentricFileSelector",
    "AssetCentricSelector",
    "AssetSubtreeSelector",
    "CanvasExternalIdSelector",
    "CanvasSelector",
    "ChartExternalIdSelector",
    "ChartOwnerSelector",
    "ChartSelector",
    "DataPointsDataSetSelector",
    "DataPointsFileSelector",
    "DataPointsSelector",
    "DataSelector",
    "DataSetSelector",
    "ExternalIdColumn",
    "FileContentSelector",
    "FileDataModelingTemplate",
    "FileDataModelingTemplateSelector",
    "FileIdentifierSelector",
    "FileMetadataTemplate",
    "FileMetadataTemplateSelector",
    "InstanceColumn",
    "InstanceFileSelector",
    "InstanceSelector",
    "InstanceSpaceSelector",
    "InstanceViewSelector",
    "InternalIdColumn",
    "RawTableSelector",
    "RecordContainerSelector",
    "SelectedContainer",
    "SelectedStream",
    "SelectedTable",
    "SelectedView",
    "Selector",
    "SelectorAdapter",
    "ThreeDModelIdSelector",
    "ThreeDSelector",
    "TimeSeriesColumn",
    "load_selector",
]



from collections.abc import Sequence
from typing import Any

from cognite_toolkit._cdf_tk.client.cdf_client import CDFResourceAPI, PagedResponse
from cognite_toolkit._cdf_tk.client.cdf_client.api import Endpoint
from cognite_toolkit._cdf_tk.client.http_client import HTTPClient, ItemsSuccessResponse, RequestMessage, SuccessResponse
from cognite_toolkit._cdf_tk.client.resource_classes.chart import (
    ChartRequest,
    ChartResponse,
    Visibility,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId


class ChartsAPI(CDFResourceAPI[ChartResponse]):
    def __init__(self, http_client: HTTPClient) -> None:
        super().__init__(
            http_client=http_client,
            method_endpoint_map={
                "create": Endpoint(
                    method="PUT", path="/storage/charts/charts", item_limit=1000, concurrency_max_workers=1
                ),
                "retrieve": Endpoint(
                    method="POST", path="/storage/charts/charts/byids", item_limit=1000, concurrency_max_workers=1
                ),
                "delete": Endpoint(
                    method="POST", path="/storage/charts/charts/delete", item_limit=1000, concurrency_max_workers=1
                ),
                "list": Endpoint(method="POST", path="/storage/charts/charts/list", item_limit=1000),
            },
        )

    def _make_url(self, path: str = "") -> str:
        return self._http_client.config.create_app_url(path)

    def _validate_page_response(self, response: SuccessResponse | ItemsSuccessResponse) -> PagedResponse[ChartResponse]:
        return PagedResponse[ChartResponse].model_validate_json(response.body)

    def create(self, items: Sequence[ChartRequest]) -> list[ChartResponse]:
        """Create charts in CDF.

        Args:
            items: List of chart Request objects to create.
        Returns:
            List of created chart Response objects.
        """
        return self._request_item_response(items, "create")

    def retrieve(self, items: Sequence[ExternalId]) -> list[ChartResponse]:
        """Retrieve charts from CDF by external ID.

        Args:
            items: List of ExternalId objects to retrieve.
        Returns:
            List of retrieved chart Response objects.
        """
        return self._request_item_response(items, "retrieve")

    def delete(self, items: Sequence[ExternalId]) -> None:
        """Delete charts from CDF by external ID.

        Args:
            items: List of ExternalId objects to delete.
        """
        self._request_no_response(items, "delete")

    def list(self, visibility: Visibility | None = None, is_owned: bool | None = None) -> list[ChartResponse]:
        """List charts based on visibility and ownership.

        Args:
            visibility: Visibility of the charts to list, either 'PUBLIC' or 'PRIVATE'.
            is_owned: Whether to list only owned charts.
        Returns:
            List of chart Response objects matching the criteria.
        """
        filter_params: dict[str, Any] = {}
        if visibility is not None:
            filter_params["visibility"] = visibility
        if is_owned is not None:
            filter_params["isOwned"] = is_owned
        body: dict[str, Any] = {}
        if filter_params:
            body["filter"] = filter_params
        endpoint = self._method_endpoint_map["list"]
        # Note that even though the internal docs specify that limit is supported for this endpoint,
        # you get: "Encountered an unknown key 'limit' at offset 50 at path: $" if you pass it.
        response = self._http_client.request_single_retries(
            RequestMessage(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                body_content=body,
            )
        ).get_success_or_raise()
        return self._validate_page_response(response).items



"""GroupCapability wrapper for Group capabilities.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Groups/operation/createGroups
"""

from typing import Any

from pydantic import model_serializer, model_validator
from pydantic_core.core_schema import FieldSerializationInfo

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject
from cognite_toolkit._cdf_tk.client.resource_classes.group._constants import ACL_NAME

from .acls import AclType


class ProjectUrlNames(BaseModelObject):
    """Project URL names for cross-project capabilities."""

    url_names: list[str]


class GroupCapability(BaseModelObject):
    """A single capability entry containing an ACL and optional project URL names."""

    acl: AclType
    project_url_names: ProjectUrlNames | None = None

    @model_validator(mode="before")
    @classmethod
    def move_acl_name(cls, value: Any) -> Any:
        """Move ACL key (e.g., 'assetsAcl') to 'acl' field for API compatibility."""
        if not isinstance(value, dict):
            return value
        if "acl" in value:
            return value
        acl_name = next((key for key in value if key.endswith("Acl")), None)
        if acl_name is None:
            return value
        value_copy = value.copy()
        acl_data = dict(value_copy.pop(acl_name))
        acl_data[ACL_NAME] = acl_name
        value_copy["acl"] = acl_data
        return value_copy

    # MyPy complains that info; FieldSerializationInfo is not compatible with info: Any
    # It is.
    @model_serializer  # type: ignore[type-var]
    def serialize_acl_name(self, info: FieldSerializationInfo) -> dict[str, Any]:
        """Serialize 'acl' field back to its specific ACL key (e.g., 'assetsAcl') for API compatibility."""
        acl_data = self.acl.model_dump(**vars(info))
        output: dict[str, Any] = {self.acl.acl_name: acl_data}
        if self.project_url_names is not None:
            output["projectUrlNames"] = self.project_url_names.model_dump(**vars(info))
        return output



from typing import Any, ClassVar, Literal

from cognite_toolkit._cdf_tk.client._resource_base import (
    BaseModelObject,
    ResponseResource,
    UpdatableRequestResource,
)

from .identifiers import ExternalId


class DataProduct(BaseModelObject):
    """Represents a data product in CDF."""

    external_id: str
    name: str

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class DataProductRequest(DataProduct, UpdatableRequestResource):
    """Request resource for creating/updating data products."""

    container_fields: ClassVar[frozenset[str]] = frozenset({"tags"})

    description: str | None = None
    schema_space: str | None = None
    is_governed: bool = False
    tags: list[str] | None = None

    def as_update(self, mode: Literal["patch", "replace"]) -> dict[str, Any]:
        update_item = super().as_update(mode)
        # schemaSpace is immutable after creation — the API rejects it in updates.
        update_item.get("update", {}).pop("schemaSpace", None)
        return update_item


class DataProductResponse(DataProduct, ResponseResource[DataProductRequest]):
    """Response resource for data products."""

    schema_space: str
    is_governed: bool
    tags: list[str]
    domains: list[str]
    description: str | None = None
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[DataProductRequest]:
        return DataProductRequest



from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from functools import cached_property, lru_cache, partial
from typing import ClassVar, Literal, TypeVar, get_args, overload

import questionary
from cognite.client.data_classes import (
    Asset,
    UserProfileList,
    filters,
)
from cognite.client.data_classes.aggregations import Count
from cognite.client.data_classes.data_modeling import ViewId
from cognite.client.data_classes.data_modeling.statistics import SpaceStatistics
from cognite.client.utils import ms_to_datetime
from questionary import Choice
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.request_classes.filters import DataModelFilter, ViewFilter
from cognite_toolkit._cdf_tk.client.resource_classes.apm_config_v1 import APMConfigResponse
from cognite_toolkit._cdf_tk.client.resource_classes.chart import ChartResponse, Visibility
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    ContainerReference,
    DataModelReference,
    DataModelResponse,
    DataModelResponseWithViews,
    SpaceResponse,
    ViewReference,
    ViewResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.dataset import DataSetResponse
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import ExternalId, RawTableId
from cognite_toolkit._cdf_tk.client.resource_classes.legacy.canvas import Canvas
from cognite_toolkit._cdf_tk.client.resource_classes.resource_view_mapping import ResourceViewMappingResponse
from cognite_toolkit._cdf_tk.client.resource_classes.three_d import ThreeDModelClassicResponse
from cognite_toolkit._cdf_tk.exceptions import ToolkitMissingResourceError, ToolkitValueError

from . import humanize_collection
from .aggregators import (
    AssetAggregator,
    AssetCentricAggregator,
    EventAggregator,
    FileAggregator,
    TimeSeriesAggregator,
)
from .useful_types import AssetCentricDestinationType

T_Type = TypeVar("T_Type", bound=Asset | DataSetResponse)


class AssetCentricInteractiveSelect(ABC):
    def __init__(self, client: ToolkitClient, operation: str) -> None:
        self.client = client
        self.operation = operation
        self._aggregator = self._get_aggregator(client)

    @abstractmethod
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        raise NotImplementedError()

    @lru_cache
    def aggregate_count(self, hierarchies: tuple[str, ...], data_sets: tuple[str, ...]) -> int:
        return self._aggregate_count(list(hierarchies), list(data_sets))

    def _aggregate_count(self, hierarchies: list[str], data_sets: list[str]) -> int:
        return self._aggregator.count(hierarchies, data_sets)

    @lru_cache
    def _get_available_data_sets(self, hierarchy: str | None = None) -> list[DataSetResponse]:
        if hierarchy is not None:
            datasets = self._aggregator.used_data_sets(hierarchy)
            return self.client.tool.datasets.retrieve(ExternalId.from_external_ids(datasets), ignore_unknown_ids=True)
        else:
            return self.client.tool.datasets.list(limit=None)

    @lru_cache
    def _get_available_hierarchies(self, data_set: str | None = None) -> list[Asset]:
        data_set_external_ids = [data_set] if data_set else None
        return list(self.client.assets.list(root=True, limit=-1, data_set_external_ids=data_set_external_ids))

    def _create_choice(self, item: Asset | DataSetResponse) -> tuple[questionary.Choice, int]:
        """Create a questionary choice for the given item."""
        if item.external_id is None:
            item_count = -1  # No count available for DataSet/Assets without external_id
        elif isinstance(item, DataSetResponse):
            item_count = self.aggregate_count(tuple(), (item.external_id,))
        elif isinstance(item, Asset):
            item_count = self.aggregate_count((item.external_id,), tuple())
        else:
            raise ToolkitValueError(f"Unexpected item type: {type(item)}. Expected Asset or DataSet.")

        return questionary.Choice(
            title=f"{item.name} ({item.external_id}) [{item_count:,}]"
            if item.name != item.external_id
            else f"({item.external_id}) [{item_count:,}]",
            value=item.external_id,
        ), item_count

    @overload
    def select_hierarchy(self, allow_empty: Literal[False] = False) -> str: ...

    @overload
    def select_hierarchy(self, allow_empty: Literal[True]) -> str | None: ...

    def select_hierarchy(self, allow_empty: Literal[False, True] = False) -> str | None:
        """Select a hierarchy interactively."""
        options = self._get_available_hierarchies()
        return self._select("hierarchy", options, allow_empty, "single")

    def select_hierarchies(self) -> list[str]:
        options = self._get_available_hierarchies()
        return self._select("hierarchies", options, False, "multiple")

    @overload
    def select_data_set(self, allow_empty: Literal[False] = False) -> str: ...

    @overload
    def select_data_set(self, allow_empty: Literal[True]) -> str | None: ...

    def select_data_set(self, allow_empty: Literal[False, True] = False) -> str | None:
        """Select a data set interactively."""
        options = self._get_available_data_sets()
        return self._select("data set", options, allow_empty, "single")

    def select_data_sets(self) -> list[str]:
        """Select multiple data sets interactively."""
        options = self._get_available_data_sets()
        return self._select("data sets", options, False, "multiple")

    def select_hierarchies_and_data_sets(self) -> tuple[list[str], list[str]]:
        what = self.select_hierarchies_or_data_sets()
        if what == "Hierarchy":
            hierarchy = self._select("hierarchy", self._get_available_hierarchies(), False, "single")
            data_sets = self._get_available_data_sets(hierarchy=hierarchy)
            if not data_sets:
                return [hierarchy], []
            selected_data_sets = self._select(f"data sets in hierarchy {hierarchy!r}", data_sets, True, "multiple")
            return [hierarchy], selected_data_sets or []
        elif what == "Data Set":
            data_set = self._select("data Set", self._get_available_data_sets(), False, "single")
            hierarchies = self._get_available_hierarchies(data_set=data_set)
            if not hierarchies:
                return [], [data_set]
            selected_hierarchies = self._select(f"hierarchies in data set {data_set!r} ", hierarchies, True, "multiple")
            return selected_hierarchies or [], [data_set]

    def select_hierarchies_or_data_sets(self) -> Literal["Hierarchy", "Data Set"]:
        what = questionary.select(
            f"Do you want to {self.operation} a hierarchy or a data set?", choices=["Hierarchy", "Data Set"]
        ).unsafe_ask()
        if what not in ["Hierarchy", "Data Set"]:
            raise ToolkitValueError(f"Unexpected selection: {what}. Aborting.")
        return what

    @overload
    def _select(
        self,
        what: str,
        options: Sequence[T_Type],
        allow_empty: Literal[False],
        plurality: Literal["single"],
    ) -> str: ...

    @overload
    def _select(
        self,
        what: str,
        options: Sequence[T_Type],
        allow_empty: Literal[True],
        plurality: Literal["single"],
    ) -> str | None: ...

    @overload
    def _select(
        self,
        what: str,
        options: Sequence[T_Type],
        allow_empty: Literal[False],
        plurality: Literal["multiple"],
    ) -> list[str]: ...

    @overload
    def _select(
        self,
        what: str,
        options: Sequence[T_Type],
        allow_empty: Literal[True],
        plurality: Literal["multiple"],
    ) -> list[str] | None: ...

    def _select(
        self,
        what: str,
        options: Sequence[Asset | DataSetResponse],
        allow_empty: Literal[True, False],
        plurality: Literal["single", "multiple"],
    ) -> str | list[str] | None:
        """Select a single item interactively."""
        if not options and not allow_empty:
            raise ToolkitValueError(f"No {what} available to select.")

        choices: list[questionary.Choice] = []
        none_sentinel = object()
        if allow_empty:
            choices.append(questionary.Choice(title=f"All {what}", value=none_sentinel))

        for choice, count in sorted(
            (self._create_choice(item) for item in options), key=lambda x: (-x[1], x[0].title)
        ):  # cont, choice.title
            if count > 0:
                choices.append(choice)
        if not choices and not allow_empty:
            raise ToolkitValueError(f"No {what} available with data to select.")

        message = f"Select a {what} to {self.operation} listed as 'name (external_id) [count]'"
        if plurality == "multiple":
            selected = questionary.checkbox(
                message,
                choices=choices,
                validate=lambda choices: "You must select at least one." if not choices else True,
            ).unsafe_ask()
        else:
            selected = questionary.select(message, choices=choices).unsafe_ask()
        if selected is None:
            raise ToolkitValueError(f"No {what} selected. Aborting.")
        elif selected is none_sentinel or (isinstance(selected, list) and none_sentinel in selected):
            return None
        return selected


class AssetInteractiveSelect(AssetCentricInteractiveSelect):
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        return AssetAggregator(self.client)


class FileMetadataInteractiveSelect(AssetCentricInteractiveSelect):
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        return FileAggregator(self.client)


class TimeSeriesInteractiveSelect(AssetCentricInteractiveSelect):
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        return TimeSeriesAggregator(self.client)


class EventInteractiveSelect(AssetCentricInteractiveSelect):
    def _get_aggregator(self, client: ToolkitClient) -> AssetCentricAggregator:
        return EventAggregator(self.client)


class RawTableInteractiveSelect:
    def __init__(self, client: ToolkitClient, operation: str) -> None:
        self.client = client
        self.operation = operation

    def _available_databases(self) -> list[str]:
        databases = self.client.tool.raw.databases.list(limit=None)
        return [db.name for db in databases]

    def _available_tables(self, database: str) -> list[RawTableId]:
        tables = self.client.tool.raw.tables.list(db_name=database, limit=None)
        return [table.as_id() for table in tables]

    def select_tables(self, database: str | None = None) -> list[RawTableId]:
        """Interactively select raw tables."""
        databases = self._available_databases()
        if not databases:
            raise ToolkitValueError("No raw databases available. Aborting.")
        if database and database not in databases:
            raise ToolkitValueError(
                f"Database '{database}' not found in available raw databases: {databases}. Aborting."
            )
        elif database:
            selected_database = database
        else:
            selected_database = questionary.select(
                f"Select a Raw Database to {self.operation}",
                choices=[questionary.Choice(title=db, value=db) for db in databases],
            ).unsafe_ask()
        available_tables = self._available_tables(selected_database)
        if not available_tables:
            raise ToolkitValueError(f"No raw tables available in database '{selected_database}'. Aborting.")

        selected_tables = questionary.checkbox(
            f"Select Raw Tables in {selected_database} to {self.operation}",
            choices=[questionary.Choice(title=f"{table.name}", value=table) for table in available_tables],
            validate=lambda choices: True if choices else "You must select at least one table.",
        ).unsafe_ask()

        return selected_tables


@dataclass
class CanvasFilter:
    visibility: Literal["public", "private"] | None = None
    created_by: Literal["user"] | None = None
    select_all: bool = False

    def as_dms_filter(self) -> filters.Filter:
        canvas_id = Canvas.get_source()
        leaf_filters: list[filters.Filter] = [
            filters.Not(filters.Equals(canvas_id.as_property_ref("isArchived"), True)),
            # When sourceCanvasId is not set, we get the newest version of the canvas
            filters.Not(filters.Exists(canvas_id.as_property_ref("sourceCanvasId"))),
        ]
        if self.visibility is not None:
            leaf_filters.append(filters.Equals(canvas_id.as_property_ref("visibility"), self.visibility))

        return filters.And(*leaf_filters)


class InteractiveCanvasSelect:
    opening_choices: ClassVar[list[questionary.Choice]] = [
        questionary.Choice(title="All public Canvases", value=CanvasFilter(visibility="public", select_all=True)),
        questionary.Choice(title="Selected public Canvases", value=CanvasFilter(visibility="public", select_all=False)),
        questionary.Choice(title="All by given user", value=CanvasFilter(created_by="user", select_all=True)),
        questionary.Choice(title="Selected by given user", value=CanvasFilter(created_by="user", select_all=False)),
        questionary.Choice(title="All Canvases", value=CanvasFilter(visibility=None, select_all=True)),
    ]

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    def select_external_ids(self) -> list[str]:
        select_filter = self._select_filter()

        return self._select_external_ids(select_filter)

    @classmethod
    def _select_filter(cls) -> CanvasFilter:
        user_response = questionary.select(
            "Which Canvases do you want to select?",
            choices=cls.opening_choices,
        ).unsafe_ask()
        return user_response

    def _select_external_ids(self, select_filter: CanvasFilter) -> list[str]:
        available_canvases = self.client.canvas.list(filter=select_filter.as_dms_filter(), limit=-1)
        if select_filter.select_all and select_filter.created_by is None:
            return [canvas.external_id for canvas in available_canvases]
        users = self.client.iam.user_profiles.list(limit=-1)
        display_name_by_user_identifier = {user.user_identifier: user.display_name or "missing" for user in users}
        if select_filter.created_by == "user":
            canvas_by_user: dict[str, list[Canvas]] = defaultdict(list)
            for canvas in available_canvases:
                canvas_by_user[canvas.created_by].append(canvas)

            user_response = questionary.select(
                "Which user do you want to select Canvases for?",
                choices=[
                    questionary.Choice(
                        title=f"{user.display_name} ({user.user_identifier})",
                        value=canvas_by_user[user.user_identifier],
                    )
                    for user in users
                    if user.user_identifier in canvas_by_user
                ],
            ).unsafe_ask()
            available_canvases = user_response

        if select_filter.select_all:
            return [canvas.external_id for canvas in available_canvases]

        if not available_canvases:
            raise ToolkitValueError("No Canvases available to select. Aborting.")

        selected_canvases = questionary.checkbox(
            "Select Canvases",
            choices=[
                questionary.Choice(
                    title=f"{canvas.name} (Created by {display_name_by_user_identifier[canvas.created_by]!r}, last updated {canvas.updated_at!r})",
                    value=canvas.external_id,
                )
                for canvas in available_canvases
            ],
            validate=lambda selected: "You must select at least one Canvas." if not selected else True,
        ).unsafe_ask()

        return selected_canvases or []


@dataclass
class ChartFilter:
    visibility: Visibility | None = None
    owned_by: Literal["user"] | None = None
    select_all: bool = False


class InteractiveChartSelect:
    opening_choices: ClassVar[list[questionary.Choice]] = [
        questionary.Choice(title="All public Charts", value=ChartFilter(visibility="PUBLIC", select_all=True)),
        questionary.Choice(title="Selected public Charts", value=ChartFilter(visibility="PUBLIC", select_all=False)),
        questionary.Choice(title="All owned by given user", value=ChartFilter(owned_by="user", select_all=True)),
        questionary.Choice(title="Selected owned by given user", value=ChartFilter(owned_by="user", select_all=False)),
    ]

    def __init__(self, client: ToolkitClient) -> None:
        self.client = client

    def select_external_ids(self) -> list[str]:
        select_filter = self._select_filter()
        return self._select_external_ids(select_filter)

    @classmethod
    def _select_filter(cls) -> ChartFilter:
        user_response = questionary.select(
            "Which Charts do you want to select?",
            choices=cls.opening_choices,
        ).unsafe_ask()
        return user_response

    def _select_external_ids(self, select_filter: ChartFilter) -> list[str]:
        available_charts = self.client.charts.list(visibility=(select_filter.visibility or "PUBLIC"))
        if select_filter.select_all and select_filter.owned_by is None:
            return [chart.external_id for chart in available_charts]

        users = self.client.iam.user_profiles.list(limit=-1)
        display_name_by_user_identifier = {
            user.user_identifier: user.display_name for user in users if user.display_name
        }
        if select_filter.owned_by == "user":
            available_charts = self._select_charts_by_user(available_charts, users)
        if select_filter.select_all:
            return [chart.external_id for chart in available_charts]
        selected_charts = questionary.checkbox(
            "Select Charts",
            choices=[
                questionary.Choice(
                    title=f"{chart.data.name} (Created by "
                    f"{display_name_by_user_identifier.get(chart.owner_id, 'missing')!r} "
                    f"- {ms_to_datetime(chart.last_updated_time)})",
                    value=chart.external_id,
                )
                for chart in available_charts
            ],
            validate=lambda selected: "You must select at least one Chart." if not selected else True,
        ).unsafe_ask()
        return selected_charts or []

    @classmethod
    def _select_charts_by_user(
        cls, available_charts: list[ChartResponse], users: UserProfileList
    ) -> list[ChartResponse]:
        chart_by_user: dict[str, list[ChartResponse]] = defaultdict(list)
        for chart in available_charts:
            chart_by_user[chart.owner_id].append(chart)
        user_response = questionary.select(
            "Which user do you want to select Charts for?",
            choices=[
                questionary.Choice(
                    title=f"{user.display_name} ({user.user_identifier})",
                    value=chart_by_user[user.user_identifier],
                )
                for user in users
                if user.user_identifier in chart_by_user
            ],
        ).unsafe_ask()
        return list(user_response)


class AssetCentricDestinationSelect:
    valid_destinations: tuple[str, ...] = get_args(AssetCentricDestinationType)

    @classmethod
    def validate(cls, destination_type: str) -> AssetCentricDestinationType:
        if destination_type not in cls.valid_destinations:
            raise ToolkitValueError(
                f"Invalid destination type: {destination_type!r}. Must be one of {humanize_collection(cls.valid_destinations)}."
            )
        # We validated the destination type above
        return destination_type  # type: ignore[return-value]

    @classmethod
    def select(cls) -> AssetCentricDestinationType:
        """Interactively select a destination type."""
        destination_type = questionary.select(
            "Select a destination type",
            choices=[questionary.Choice(title=dest.capitalize(), value=dest) for dest in cls.valid_destinations],
        ).unsafe_ask()
        # We only input valid destination types, so we can safely skip MyPy's type checking here
        return destination_type

    @classmethod
    def get(cls, destination_type: str | None = None) -> AssetCentricDestinationType:
        """Get the destination type, either from the input or interactively."""
        if destination_type is None:
            return cls.select()
        return cls.validate(destination_type)


@dataclass
class ViewSelectFilter:
    """Filter for creating view options when selecting views for instance selection.

    This filter is used to narrow down the list of views presented to the user when they are asked to select a view.

    Args:
        strategy: The strategy used to create the view options.
            - "SchemaSpace": Ask the user to select a schema space first.
        include_global: Whether to include global views in the options.
        schema_space : The schema space to look for view in.
        instance_type: If set, only include views that can be used for the given instance type ("node", "edge", or "all").
        mapped_container: If set, only include views that have the given container in their mapped containers.

    """

    strategy: Literal["schemaSpace", "dataModel"] = "schemaSpace"
    include_global: bool | None = None
    schema_space: str | None = None
    instance_type: Literal["node", "edge", "all"] | None = None
    mapped_container: ContainerReference | None = None
    data_model: DataModelReference | None = None

    def __str__(self) -> str:
        message: list[str] = []
        if self.include_global:
            message.append("including global")
        if self.schema_space:
            message.append(f"in schema space {self.schema_space!r}")
        if self.instance_type and self.instance_type != "all":
            message.append(f"for instance type {self.instance_type!r}")
        if self.mapped_container:
            message.append(f"with mapped container {self.mapped_container!r}")
        return f"Views {humanize_collection(message)}"


class DataModelingSelect:
    """A utility class to select Data Modeling nodes interactively."""

    def __init__(self, client: ToolkitClient, operation: str, console: Console | None = None) -> None:
        self.client = client
        self.operation = operation
        self.console = console or Console()
        self._available_spaces: list[SpaceResponse] | None = None

    @cached_property
    def stats_by_space(self) -> dict[str, SpaceStatistics]:
        result = self.client.data_modeling.statistics.spaces.list()
        return {stat.space: stat for stat in result}

    @overload
    def select_view(
        self,
        multiselect: Literal[False] = False,
        message: str | None = None,
        filter: ViewSelectFilter | None = None,
    ) -> ViewResponse: ...

    @overload
    def select_view(
        self,
        multiselect: Literal[True],
        message: str | None = None,
        filter: ViewSelectFilter | None = None,
    ) -> list[ViewResponse]: ...

    def select_view(
        self,
        multiselect: bool = False,
        message: str | None = None,
        filter: ViewSelectFilter | None = None,
    ) -> ViewResponse | list[ViewResponse]:
        """Select one or more views interactively.

        Args:
            multiselect: Whether to allow selecting multiple views.
             message: The message to display when prompting for a view. If None, a default message
                will be used.
            filter: Optional filter to apply when listing views. If None, no additional filtering will be applied.
        Returns:
            The selected view(s).

        Raises:
            ToolkitValueError: If no view(s) are selected or if the selected view(s) are not valid.
            ToolkitMissingResourceError: If no views are found in the selected space.

        """
        views = self._get_view_options(filter or ViewSelectFilter())
        question = message or f"Which view do you want to use to select instances to {self.operation}?"
        choices = [
            Choice(title=f"{view.space}:{view.external_id}(version={view.version})", value=view) for view in views
        ]
        if multiselect:
            selected_views = questionary.checkbox(
                question,
                choices=choices,
                validate=lambda choises: True if choises else "You must select at least one view.",
            ).unsafe_ask()
            if not isinstance(selected_views, list) or not all(isinstance(v, ViewResponse) for v in selected_views):
                raise ToolkitValueError(f"Selected views is not a valid list of View objects: {selected_views!r}")
            return selected_views
        else:
            selected_views = questionary.select(question, choices=choices).unsafe_ask()
            if not isinstance(selected_views, ViewResponse):
                raise ToolkitValueError(f"Selected view is not a valid View object: {selected_views!r}")
            return selected_views

    def _get_view_options(self, filter: ViewSelectFilter) -> list[ViewResponse]:
        if filter.strategy == "schemaSpace":
            selected_space = (
                filter.schema_space
                or self.select_schema_space(
                    filter.include_global or False,
                    message=f"In which Spaces is the view you will use to select instances to {self.operation}?",
                ).space
            )
            views = self.client.tool.views.list(
                filter=ViewFilter(
                    space=selected_space,
                    include_inherited_properties=True,
                    include_global=filter.include_global,
                ),
                limit=None,
            )
        elif filter.strategy == "dataModel":
            if filter.data_model:
                data_models = self.client.tool.data_models.retrieve([filter.data_model], inline_views=True)
                if not data_models or not data_models[0].views:
                    raise ToolkitMissingResourceError(
                        f"No views found in data model {filter.data_model!r} for filter: {filter!s}"
                    )
                datamodel = data_models[0]
            else:
                datamodel = self.select_data_model(
                    inline_views=True,
                    message=f"Select the data model through which to {self.operation}:",
                    schema_space=filter.schema_space,
                    include_global=filter.include_global,
                )
            views = datamodel.views or []
            parents = {parent for view in views for parent in view.implements or []}
            # We only allow the user to select child views
            views = [view for view in views if view.as_id() not in parents]
        else:
            raise NotImplementedError(f"Strategy {filter.strategy} is not implemented.")
        views = [
            view
            for view in views
            if filter.instance_type in (None, "all", view.used_for)
            and (filter.mapped_container is None or filter.mapped_container in view.mapped_containers)
        ]
        if not views:
            raise ToolkitMissingResourceError(f"No views found that match the filter: {filter!s}")
        return views

    @overload
    def select_data_model(
        self,
        inline_views: Literal[False] = False,
        message: str | None = None,
        schema_space: str | None = None,
        include_global: bool | None = None,
    ) -> DataModelResponse: ...

    @overload
    def select_data_model(
        self,
        inline_views: Literal[True],
        message: str | None = None,
        schema_space: str | None = None,
        include_global: bool | None = None,
    ) -> DataModelResponseWithViews: ...

    def select_data_model(
        self,
        inline_views: Literal[True, False] = False,
        message: str | None = None,
        schema_space: str | None = None,
        include_global: bool | None = None,
    ) -> DataModelResponse | DataModelResponseWithViews:
        datamodels = self.client.tool.data_models.list(
            inline_views=inline_views,
            filter=DataModelFilter(space=schema_space, all_versions=False, include_global=include_global),
            limit=None,
        )
        if not datamodels:
            raise ToolkitMissingResourceError(
                f"No data models found {f'in schema space {schema_space!r}' if schema_space else ''}."
            )

        message = message or f"Select a data model to {self.operation}:"
        selected_datamodel = questionary.select(
            message,
            choices=[
                Choice(
                    title=f"{dm.space}:{dm.external_id} (version={dm.version})",
                    value=dm,
                )
                for dm in datamodels
            ],
        ).unsafe_ask()
        if not isinstance(selected_datamodel, DataModelResponse | DataModelResponseWithViews):
            raise ToolkitValueError(
                f"Selected data model is not a valid DataModelResponse object: {selected_datamodel!r}"
            )
        return selected_datamodel

    def select_schema_space(self, include_global: bool, message: str | None = None) -> SpaceResponse:
        message = message or f"Select the space to {self.operation}:"
        spaces = self._get_available_spaces(include_global)
        schema_spaces = {
            space
            for space, stats in self.stats_by_space.items()
            if (stats.containers + stats.views + stats.data_models) > 0
        }
        spaces = [space for space in spaces if space.space in schema_spaces or space.is_global]
        if not spaces:
            raise ToolkitMissingResourceError("No spaces with schema (containers, views, or data models) found.")
        selected_space = questionary.select(
            message,
            [Choice(title=space.space, value=space) for space in sorted(spaces, key=lambda s: s.space)],
        ).unsafe_ask()
        if not isinstance(selected_space, SpaceResponse):
            raise ToolkitValueError(f"Selected space is not a valid Space object: {selected_space!r}")
        return selected_space

    def select_instance_type(
        self, view_used_for: Literal["node", "edge", "all"] | None = None, message: str | None = None
    ) -> Literal["node", "edge"]:
        """Selects an instance type (node or edge) interactively.

        Args:
            view_used_for: If 'node' or 'edge', that type is returned directly.
                           If 'all' or None, the user is prompted to select.

        Returns:
            The selected instance type, either 'node' or 'edge'.
        """
        if view_used_for is not None and view_used_for != "all":
            return view_used_for
        selected_instance_type = questionary.select(
            message or f"What type of instances do you want to {self.operation}?",
            choices=[
                Choice(title="Nodes", value="node"),
                Choice(title="Edges", value="edge"),
            ],
        ).unsafe_ask()

        if selected_instance_type not in ("node", "edge"):
            raise ToolkitValueError(f"Selected instance type is not valid: {selected_instance_type!r}")
        return selected_instance_type

    def select_space_type(self) -> Literal["instance", "schema", "empty"]:
        selected_space_type = questionary.select(
            f"What type of spaces do you want to {self.operation}?",
            choices=[
                Choice(title="Instance spaces with instances", value="instance"),
                Choice(title="Schema spaces with schema", value="schema"),
                Choice(title="Empty spaces without schema or instances", value="empty"),
            ],
        ).unsafe_ask()
        if selected_space_type not in ["instance", "schema", "empty"]:
            raise ToolkitValueError(f"Selected space type is not valid: {selected_space_type!r}")
        return selected_space_type

    @overload
    def select_instance_space(
        self,
        multiselect: Literal[False],
        selected_view: ViewReference | None = None,
        instance_type: Literal["node", "edge"] | None = None,
        message: str | None = None,
        include_empty: bool = False,
    ) -> str: ...

    @overload
    def select_instance_space(
        self,
        multiselect: Literal[True] = True,
        selected_view: ViewReference | None = None,
        instance_type: Literal["node", "edge"] | None = None,
        message: str | None = None,
        include_empty: bool = False,
    ) -> list[str]: ...

    def select_instance_space(
        self,
        multiselect: bool = True,
        selected_view: ViewReference | None = None,
        instance_type: Literal["node", "edge"] | None = None,
        message: str | None = None,
        include_empty: bool = False,
    ) -> str | list[str]:
        if (selected_view is not None and instance_type is None) or (
            selected_view is None and instance_type is not None
        ):
            raise ToolkitValueError("Both selected_view and instance_type must be provided or both must be None.")
        if selected_view is not None and instance_type is not None:
            all_spaces = self._get_available_spaces(include_global=False)
            count_by_space = self._get_instance_count_in_view_by_space(all_spaces, selected_view, instance_type)
        else:
            count_by_space = {
                space: stats.nodes + stats.edges
                for space, stats in self.stats_by_space.items()
                if (stats.nodes + stats.edges) > 0
                or (
                    include_empty
                    and (stats.nodes + stats.edges + stats.containers + stats.views + stats.data_models) == 0
                )
            }

        if not count_by_space:
            raise ToolkitMissingResourceError(
                f"No instances found in any space for the view {selected_view!r} with instance type {instance_type!r}."
            )
        if len(count_by_space) == 1:
            selected_spaces = next(iter(count_by_space.keys()))
            self.console.print(f"Only one space with instances found: {selected_spaces!r}. Using this space.")
            return [selected_spaces] if multiselect else selected_spaces

        if not message:
            message = f"In which Space{'(s)' if multiselect else ''} do you want to {self.operation} instances?"
        choices = [
            Choice(title=f"{space} ({count:,} instances)", value=space)
            for space, count in sorted(count_by_space.items(), key=lambda item: item[1], reverse=True)
        ]
        if multiselect:
            selected_spaces = questionary.checkbox(
                message,
                choices=choices,
                validate=lambda choices: True if choices else "You must select at least one space.",
            ).unsafe_ask()
        else:
            selected_spaces = questionary.select(message, choices=choices).unsafe_ask()
        if selected_spaces is None or (isinstance(selected_spaces, list) and len(selected_spaces) == 0):
            raise ToolkitValueError(
                f"No instance space selected to {self.operation}. Use arrow keys to navigate and space key to select. Press enter to confirm."
            )
        return selected_spaces

    @overload
    def select_empty_spaces(self, multiselect: Literal[False]) -> str: ...

    @overload
    def select_empty_spaces(self, multiselect: Literal[True]) -> list[str]: ...

    def select_empty_spaces(self, multiselect: bool = True) -> str | list[str]:
        empty_spaces = [
            space
            for space, stats in self.stats_by_space.items()
            if (stats.nodes + stats.edges + stats.containers + stats.views + stats.data_models) == 0
        ]
        if not empty_spaces:
            raise ToolkitMissingResourceError("No empty spaces found.")
        if len(empty_spaces) == 1:
            selected_space = empty_spaces[0]
            self.console.print(f"Only one empty space found: {selected_space!r}. Using this space.")
            return [selected_space] if multiselect else selected_space

        message = f"In which empty Space{'(s)' if multiselect else ''} do you want to {self.operation}?"
        choices = [Choice(title=f"{space}", value=space) for space in sorted(empty_spaces)]
        if multiselect:
            selected_spaces = questionary.checkbox(
                message,
                choices=choices,
                validate=lambda choises: True if choises else "You must select at least one space.",
            ).unsafe_ask()
        else:
            selected_spaces = questionary.select(message, choices=choices).unsafe_ask()
        if selected_spaces is None or (isinstance(selected_spaces, list) and len(selected_spaces) == 0):
            raise ToolkitValueError(f"No empty space selected to {self.operation}. Aborting.")
        return selected_spaces

    def _get_instance_count_in_view_by_space(
        self, all_spaces: list[SpaceResponse], view_id: ViewReference, instance_type: Literal["node", "edge"]
    ) -> dict[str, float]:
        count_by_space: dict[str, float] = {}
        result = self.client.data_modeling.statistics.project()
        read_limit = result.concurrent_read_limit

        with ThreadPoolExecutor(max_workers=read_limit // 2) as executor:
            results = executor.map(
                partial(self._instance_count_space, view_id=view_id, instance_type=instance_type),
                (space.space for space in all_spaces),
            )
            for space_name, count in results:
                if count > 0:
                    count_by_space[space_name] = count

        return count_by_space

    @lru_cache
    def _instance_count_space(
        self, space: str, view_id: ViewReference, instance_type: Literal["node", "edge"]
    ) -> tuple[str, float]:
        """Get the count of instances in a specific space for a given view and instance type."""
        sdk_view_id = ViewId(space=view_id.space, external_id=view_id.external_id, version=view_id.version)
        return space, self.client.data_modeling.instances.aggregate(
            sdk_view_id, Count("externalId"), instance_type=instance_type, space=space
        ).value or 0.0

    def _get_available_spaces(self, include_global: bool = False) -> list[SpaceResponse]:
        if self._available_spaces is None:
            self._available_spaces = self.client.tool.spaces.list(include_global=True)
        if include_global:
            return self._available_spaces
        return [space for space in self._available_spaces if not space.is_global]


class ResourceViewMappingInteractiveSelect:
    def __init__(self, client: ToolkitClient, operation: str) -> None:
        self.client = client
        self.operation = operation

    def select_resource_view_mapping(self, resource_type: str) -> ResourceViewMappingResponse:
        """Select a Resource View Mapping interactively.

        Args:
            resource_type: The resource type to filter Resource View Mappings by.
        Returns:
            The selected Resource View Mapping.
        """
        mappings = self.client.migration.resource_view_mapping.list(resource_type=resource_type, limit=-1)
        if not mappings:
            raise ToolkitMissingResourceError(f"No Resource View Mappings found for resource type {resource_type!r}.")
        choices = [
            Choice(
                title=f"{mapping.external_id} ({mapping.view_id!r})",
                value=mapping,
            )
            for mapping in mappings
        ]
        selected_mapping = questionary.select(
            f"Which Resource View Mapping do you want to use to {self.operation}? [identifier] (ingestion view)",
            choices=choices,
        ).unsafe_ask()
        if not isinstance(selected_mapping, ResourceViewMappingResponse):
            raise ToolkitValueError(
                f"Selected Resource View Mapping is not a valid ResourceViewMappingResponse object: {selected_mapping!r}"
            )
        return selected_mapping


class ThreeDInteractiveSelect:
    def __init__(self, client: ToolkitClient, operation: str) -> None:
        self.client = client
        self.operation = operation

    def select_three_d_models(
        self, model_type: Literal["classic", "dm"] | None = None
    ) -> list[ThreeDModelClassicResponse]:
        """Select multiple 3D models interactively."""
        if model_type is None:
            model_type = questionary.select(
                f"What type of 3D models do you want to {self.operation}?",
                choices=[
                    Choice(title="Classic models", value="classic"),
                    Choice(title="Data modeling 3D", value="dm"),
                ],
            ).unsafe_ask()
        published = questionary.select(
            f"Do you want to {self.operation} published or unpublished 3D models?",
            choices=[
                Choice(title="Published models", value=True),
                Choice(title="Unpublished models", value=False),
                Choice(title="Both published and unpublished models", value=None),
            ],
        ).unsafe_ask()

        models = self.client.tool.three_d.models_classic.list(
            published=published, include_revision_info=True, limit=None
        )
        if model_type == "classic":
            models = [model for model in models if model.space is None]
        else:
            models = [model for model in models if model.space is not None]
        if not models:
            raise ToolkitMissingResourceError(
                f"No 3D models found for type {model_type!r} with published={published!r}."
            )

        choices = [Choice(title=f"{model.name} ({model.id})", value=model) for model in models]
        selected_models = questionary.checkbox(
            f"Select 3D models to {self.operation}:",
            choices=choices,
            validate=lambda choices: True if choices else "You must select at least one 3D model.",
        ).unsafe_ask()
        if selected_models is None or len(selected_models) == 0:
            raise ToolkitValueError("No 3D models selected.")
        return selected_models


class APMConfigInteractiveSelect:
    """Interactive select for APM Config (Infield V1) configurations."""

    def __init__(self, client: ToolkitClient, operation: str) -> None:
        self.client = client
        self.operation = operation

    def _get_choices(self) -> list[Choice]:
        """Fetch APM configs and create choices for selection.

        Returns:
            List of questionary Choice objects for the available APM configs.

        Raises:
            ToolkitMissingResourceError: If no APM configs are found.
        """
        configs = self.client.infield.apm_config.list(limit=None)
        if not configs:
            raise ToolkitMissingResourceError("No APM configs found.")

        return [
            Choice(
                title=f"{config.name} ({config.external_id})" if config.name else f"({config.external_id})",
                value=config,
            )
            for config in configs
        ]

    def select_apm_configs(self) -> list[APMConfigResponse]:
        """Select multiple APM configs interactively.

        Returns:
            List of selected APMConfigResponse objects.

        Raises:
            ToolkitMissingResourceError: If no APM configs are found.
            ToolkitValueError: If no APM configs are selected.
        """
        choices = self._get_choices()
        selected_configs = questionary.checkbox(
            f"Select APM configs to {self.operation}:",
            choices=choices,
            validate=lambda choices: True if choices else "You must select at least one APM config.",
        ).unsafe_ask()
        if selected_configs is None or len(selected_configs) == 0:
            raise ToolkitValueError("No APM configs selected.")
        return selected_configs

    def select_apm_config(self) -> APMConfigResponse:
        """Select a single APM config interactively.

        Returns:
            The selected APMConfigResponse object.

        Raises:
            ToolkitMissingResourceError: If no APM configs are found.
            ToolkitValueError: If no APM config is selected.
        """
        choices = self._get_choices()
        selected_config = questionary.select(
            f"Select an APM config to {self.operation}:",
            choices=choices,
        ).unsafe_ask()
        if selected_config is None:
            raise ToolkitValueError("No APM config selected.")
        if not isinstance(selected_config, APMConfigResponse):
            raise ToolkitValueError(f"Selected APM config is not a valid APMConfigResponse object: {selected_config!r}")
        return selected_config



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import random
import time
import warnings
from collections import defaultdict
from collections.abc import Callable, Hashable, Iterable, Sequence
from copy import deepcopy
from pathlib import Path
from typing import Any, Literal, cast, final

from cognite.client.data_classes import (
    ClientCredentials,
    OidcCredentials,
)
from cognite.client.data_classes.capabilities import (
    Capability,
    TransformationsAcl,
)
from rich import print
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.http_client import ToolkitAPIError
from cognite_toolkit._cdf_tk.client.request_classes.filters import (
    TransformationFilter,
    TransformationNotificationFilter,
)
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import (
    DataModelReference,
    SpaceReference,
    ViewReference,
)
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import (
    ExternalId,
    InternalId,
    RawDatabaseId,
    RawTableId,
    TransformationNotificationId,
)
from cognite_toolkit._cdf_tk.client.resource_classes.transformation import (
    NonceCredentials,
    TransformationRequest,
    TransformationResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.transformation_notification import (
    TransformationNotificationRequest,
    TransformationNotificationResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.transformation_schedule import (
    TransformationScheduleRequest,
    TransformationScheduleResponse,
)
from cognite_toolkit._cdf_tk.constants import BUILD_FOLDER_ENCODING
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.exceptions import (
    ResourceCreationError,
    ToolkitFileNotFoundError,
    ToolkitInvalidParameterNameError,
    ToolkitNotSupported,
    ToolkitRequiredValueError,
    ToolkitYAMLFormatError,
)
from cognite_toolkit._cdf_tk.resource_classes import (
    TransformationNotificationYAML,
    TransformationScheduleYAML,
    TransformationYAML,
)
from cognite_toolkit._cdf_tk.tk_warnings import HighSeverityWarning, MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils import (
    calculate_secure_hash,
    humanize_collection,
    in_dict,
    load_yaml_inject_variables,
    quote_int_value_by_key_in_yaml,
    safe_read,
    sanitize_filename,
)
from cognite_toolkit._cdf_tk.utils.cdf import read_auth, try_find_error
from cognite_toolkit._cdf_tk.utils.collection import chunker
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable

from .auth import GroupAllScopedCRUD
from .data_organization import DataSetsCRUD
from .datamodel import DataModelCRUD, SpaceCRUD, ViewCRUD
from .group_scoped import GroupResourceScopedCRUD
from .raw import RawDatabaseCRUD, RawTableCRUD


@final
class TransformationCRUD(ResourceCRUD[ExternalId, TransformationRequest, TransformationResponse]):
    folder_name = "transformations"
    resource_cls = TransformationResponse
    resource_write_cls = TransformationRequest
    kind = "Transformation"
    yaml_cls = TransformationYAML
    dependencies = frozenset(
        {
            DataSetsCRUD,
            RawDatabaseCRUD,
            GroupAllScopedCRUD,
            SpaceCRUD,
            ViewCRUD,
            DataModelCRUD,
            RawTableCRUD,
            RawDatabaseCRUD,
            GroupResourceScopedCRUD,
        }
    )
    _doc_url = "Transformations/operation/createTransformations"
    _hash_key = "-- cdf-auth"

    # The API supports creating/updating 1000 transformations in a single request,
    # however, when the transformation has credentials, the session API times out on
    # larger number of transformations. Thus, we use a conservative batch size.
    _BATCH_SIZE = 20  # The maximum number of transformations to create in a single batch

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None = None):
        super().__init__(client, build_dir, console)
        self._authentication_by_id_operation: dict[
            tuple[str, Literal["read", "write"]], OidcCredentials | ClientCredentials
        ] = {}

    @property
    def display_name(self) -> str:
        return "transformations"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[TransformationRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []

        actions = (
            [TransformationsAcl.Action.Read]
            if read_only
            else [TransformationsAcl.Action.Read, TransformationsAcl.Action.Write]
        )
        scope: TransformationsAcl.Scope.All | TransformationsAcl.Scope.DataSet = TransformationsAcl.Scope.All()  # type: ignore[valid-type]
        if items is not None:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = TransformationsAcl.Scope.DataSet(list(data_set_ids))

        return TransformationsAcl(actions, scope)

    @classmethod
    def get_id(cls, item: TransformationResponse | TransformationRequest | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if item.external_id is None:
            raise ToolkitRequiredValueError("Transformation must have external_id set.")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def get_internal_id(cls, item: TransformationResponse | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if item.id is None:
            raise ToolkitRequiredValueError("Transformation must have id set.")
        return item.id

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: ExternalId) -> str:
        return sanitize_filename(id.external_id)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, ExternalId(external_id=item["dataSetExternalId"])
        if destination := item.get("destination", {}):
            if not isinstance(destination, dict):
                return
            if destination.get("type") == "raw" and in_dict(("database", "table"), destination):
                yield RawDatabaseCRUD, RawDatabaseId(name=destination["database"])
                yield RawTableCRUD, RawTableId(db_name=destination["database"], name=destination["table"])
            elif destination.get("type") in ("nodes", "edges") and (view := destination.get("view", {})):
                if space := destination.get("instanceSpace"):
                    yield SpaceCRUD, SpaceReference(space=space)
                if in_dict(("space", "externalId", "version"), view):
                    view["version"] = str(view["version"])
                    yield ViewCRUD, ViewReference.model_validate(view)
            elif destination.get("type") == "instances":
                if space := destination.get("instanceSpace"):
                    yield SpaceCRUD, SpaceReference(space=space)
                if data_model := destination.get("dataModel"):
                    if in_dict(("space", "externalId", "version"), data_model):
                        data_model["version"] = str(data_model["version"])
                        yield DataModelCRUD, DataModelReference.model_validate(data_model)

    def safe_read(self, filepath: Path | str) -> str:
        # If the destination is a DataModel or a View we need to ensure that the version is a string
        return quote_int_value_by_key_in_yaml(safe_read(filepath, encoding=BUILD_FOLDER_ENCODING), key="version")

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        resources = load_yaml_inject_variables(
            self.safe_read(filepath),
            environment_variables or {},
            original_filepath=filepath,
        )

        raw_list = resources if isinstance(resources, list) else [resources]
        for item in raw_list:
            query_file: Path | None = None
            if "queryFile" in item:
                if filepath is None:
                    raise ValueError("filepath must be set if queryFile is set")
                query_file = filepath.parent / Path(item.pop("queryFile"))

            external_id = self.get_id(item).external_id
            if query_file is None and "query" not in item:
                if filepath is None:
                    raise ValueError("filepath must be set if query is not set")
                warning = HighSeverityWarning(
                    f"query property or is missing in {filepath.as_posix()!r}. It can be inline or a separate file named {filepath.stem}.sql or {external_id}.sql",
                )
                warning.print_warning(console=self.console)
            elif query_file and not query_file.exists():
                # We checked above that filepath is not None
                raise ToolkitFileNotFoundError(f"Query file {query_file.as_posix()} not found", filepath)
            elif query_file and "query" in item:
                raise ToolkitYAMLFormatError(
                    f"query property is ambiguously defined in both the yaml file and a separate file named {query_file}\n"
                    f"Please remove one of the definitions, either the query property in {filepath} or the file {query_file}",
                    filepath,
                )
            elif query_file:
                item["query"] = safe_read(query_file, encoding=BUILD_FOLDER_ENCODING)

            auth_dict: dict[str, Any] = {}
            if "authentication" in item:
                auth_dict["authentication"] = item["authentication"]
            if auth_dict:
                auth_hash = calculate_secure_hash(auth_dict, shorten=True)
                if "query" in item:
                    hash_str = f"{self._hash_key}: {auth_hash}"
                    if not item["query"].startswith(self._hash_key):
                        item["query"] = f"{hash_str}\n{item['query']}"

            if "sourceOidcCredentials" in item:
                raise ToolkitNotSupported(
                    "The property 'sourceOidcCredentials' is not supported. Use 'authentication.read' instead."
                )

            if "destinationOidcCredentials" in item:
                raise ToolkitNotSupported(
                    "The property 'destinationOidcCredentials' is not supported. Use 'authentication.write' instead."
                )

            if "sourceNonce" in item:
                raise ToolkitNotSupported(
                    "The property 'sourceNonce' is not supported by Toolkit. Use 'authentication.read' instead,"
                    "then Toolkit will dynamically set the nonce."
                )

            if "destinationNonce" in item:
                raise ToolkitNotSupported(
                    "The property 'destinationNonce' is not supported by Toolkit. Use 'authentication.write' instead,"
                    "then Toolkit will dynamically set the nonce."
                )
            auth = item.pop("authentication", None)
            if isinstance(auth, dict) and "read" in auth:
                self._authentication_by_id_operation[(external_id, "read")] = read_auth(
                    auth["read"],
                    self.client.config,
                    external_id,
                    "transformation",
                    allow_oidc=True,
                    console=self.console,
                )
            elif isinstance(auth, dict) and "write" not in auth:
                self._authentication_by_id_operation[(external_id, "read")] = read_auth(
                    auth,
                    self.client.config,
                    external_id,
                    "transformation",
                    allow_oidc=True,
                    console=self.console,
                )
            if isinstance(auth, dict) and "write" in auth:
                self._authentication_by_id_operation[(external_id, "write")] = read_auth(
                    auth["write"],
                    self.client.config,
                    external_id,
                    "transformation",
                    allow_oidc=True,
                    console=self.console,
                )
            elif isinstance(auth, dict) and "read" not in auth:
                self._authentication_by_id_operation[(external_id, "write")] = read_auth(
                    auth,
                    self.client.config,
                    external_id,
                    "transformation",
                    allow_oidc=True,
                    console=self.console,
                )
        return raw_list

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> TransformationRequest:
        invalid_parameters: dict[str, str] = {}
        if "action" in resource and "conflictMode" not in resource:
            invalid_parameters["action"] = "conflictMode"
        if "shared" in resource and "isPublic" not in resource:
            invalid_parameters["shared"] = "isPublic"
        if invalid_parameters:
            raise ToolkitInvalidParameterNameError(
                "Parameters invalid. These are specific for the "
                "'transformation-cli' and not supported by cognite-toolkit",
                resource.get("externalId", "<Missing>"),
                invalid_parameters,
            )

        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if "conflictMode" not in resource:
            resource["conflictMode"] = "upsert"
        if "ignoreNullFields" not in resource:
            resource["ignoreNullFields"] = True
        return TransformationRequest.model_validate(resource)

    def dump_resource(self, resource: TransformationResponse, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if "isPublic" in dumped and "isPublic" not in local:
            # Default set from server side.
            dumped.pop("isPublic")
        if "authentication" in local:
            # The hash added to the beginning of the query detects the change in the authentication
            dumped["authentication"] = local["authentication"]
        cdf_destination = dumped.get("destination", {})
        local_destination = local.get("destination", {})
        if isinstance(cdf_destination, dict) and isinstance(local_destination, dict):
            if cdf_destination.get("instanceSpace") is None and "instanceSpace" not in local_destination:
                cdf_destination.pop("instanceSpace", None)
        if not dumped.get("query") and "query" not in local:
            dumped.pop("query", None)
        if dumped.get("conflictMode") == "upsert" and "conflictMode" not in local:
            # Default set from server side.
            dumped.pop("conflictMode", None)
        return dumped

    def split_resource(
        self, base_filepath: Path, resource: dict[str, Any]
    ) -> Iterable[tuple[Path, dict[str, Any] | str]]:
        if query := resource.pop("query", None):
            yield base_filepath.with_suffix(".sql"), cast(str, query)

        yield base_filepath, resource

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path[-1] == "scopes":
            return diff_list_hashable(local, cdf)
        return super().diff_list(local, cdf, json_path)

    def create(self, items: Sequence[TransformationRequest]) -> list[TransformationResponse]:
        return self._execute_in_batches(items, self.client.tool.transformations.create)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[TransformationResponse]:
        return self.client.tool.transformations.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[TransformationRequest]) -> list[TransformationResponse]:
        def update(transformations: Sequence[TransformationRequest]) -> list[TransformationResponse]:
            return self.client.tool.transformations.update(list(transformations), mode="replace")

        return self._execute_in_batches(items, update)

    def _create_auth_creation_error(self, items: Sequence[TransformationRequest]) -> ResourceCreationError | None:
        hints_by_id: dict[str, list[str]] = defaultdict(list)
        for item in items:
            if not item.external_id:
                continue
            read_credentials = self._authentication_by_id_operation.get((item.external_id, "read"))
            hint_source = try_find_error(read_credentials)
            if hint_source:
                hints_by_id[item.external_id].append(hint_source)
            write_credentials = self._authentication_by_id_operation.get((item.external_id, "write"))
            if hint_dest := try_find_error(write_credentials):
                if hint_dest != hint_source:
                    hints_by_id[item.external_id].append(hint_dest)
        if hints_by_id:
            body = "\n".join(f"  {id_} - {humanize_collection(hints)}" for id_, hints in hints_by_id.items())
            return ResourceCreationError(
                f"Failed to create Transformation(s) du to likely invalid credentials:\n{body}",
            )
        return None

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.transformations.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _execute_in_batches(
        self,
        items: Sequence[TransformationRequest],
        api_call: Callable[[Sequence[TransformationRequest]], list[TransformationResponse]],
    ) -> list[TransformationResponse]:
        results: list[TransformationResponse] = []
        for chunk in chunker(items, self._BATCH_SIZE):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self._update_nonce(chunk)
            try:
                chunk_results = api_call(chunk)
            except ToolkitAPIError as e:
                if e.code in (401, 403):
                    if error := self._create_auth_creation_error(chunk):
                        raise error from e
                    raise
                elif "Failed to bind session using nonce" in e.message and len(chunk) > 1:
                    results.extend(self._execute_one_by_one(chunk, api_call))
                else:
                    raise
            else:
                results.extend(chunk_results)
        return results

    def _execute_one_by_one(
        self,
        chunk: Sequence[TransformationRequest],
        api_call: Callable[[Sequence[TransformationRequest]], list[TransformationResponse]],
    ) -> list[TransformationResponse]:
        MediumSeverityWarning(
            f"Failed to create {len(chunk)} transformations in a batch due to nonce binding error. "
            "Trying to recover by creating them one by one."
        ).print_warning(console=self.client.console)
        # Retry one by one
        failed_ids: list[str] = []
        success_count = 0
        delay = 0.3
        self._sleep_with_jitter(delay, delay + 0.3)
        results: list[TransformationResponse] = []
        for item in chunk:
            try:
                recovered = api_call([item])
            except ToolkitAPIError as e:
                if "Failed to bind session using nonce" in e.message:
                    failed_ids.append(item.external_id or "<missing>")
                    self._sleep_with_jitter(delay, delay + 0.3)
                else:
                    raise
            else:
                results.extend(recovered)
                success_count += 1
        message = f"  [bold]RECOVERY COMPLETE:[/] Successfully created {success_count:,} transformations"
        if failed_ids:
            message += f", failed to create {len(failed_ids):,} transformations: {humanize_collection(failed_ids)}"
        else:
            message += "."
        if failed_ids:
            HighSeverityWarning(message).print_warning(include_timestamp=True, console=self.client.console)
        else:
            self.client.console.print(message)
        return results

    @staticmethod
    def _sleep_with_jitter(base_delay: float, max_delay: float) -> None:
        """Sleeps for a random duration between base_delay and max_delay (inclusive)."""
        sleep_time = random.uniform(base_delay, max_delay)
        time.sleep(sleep_time)

    def _update_nonce(self, items: Sequence[TransformationRequest]) -> None:
        for item in items:
            if not item.external_id:
                raise ToolkitRequiredValueError("Transformation must have external_id set.")
            if item.source_nonce is None and (
                read_credentials := self._authentication_by_id_operation.get((item.external_id, "read"))
            ):
                item.source_nonce = self._create_nonce(read_credentials)
            if item.destination_nonce is None and (
                write_credentials := self._authentication_by_id_operation.get((item.external_id, "write"))
            ):
                item.destination_nonce = self._create_nonce(write_credentials)

    def _create_nonce(self, credentials: OidcCredentials | ClientCredentials) -> NonceCredentials:
        if isinstance(credentials, ClientCredentials):
            session = self.client.iam.sessions.create(credentials)
            nonce = NonceCredentials(
                session_id=session.id,
                nonce=session.nonce,
                cdf_project_name=self.client.config.project,
                client_id=credentials.client_id,
            )
        elif isinstance(credentials, OidcCredentials):
            config = deepcopy(self.client.config)
            config.project = credentials.cdf_project_name
            config.credentials = credentials.as_credential_provider()
            other_client = ToolkitClient(config)
            session = other_client.iam.sessions.create(credentials.as_client_credentials())
            nonce = NonceCredentials(
                session_id=session.id,
                nonce=session.nonce,
                cdf_project_name=credentials.cdf_project_name,
                client_id=credentials.client_id,
            )
        else:
            raise ValueError(f"Error in TransformationLoader: {type(credentials)} is not a valid credentials type")
        return nonce

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[TransformationResponse]:
        filter = TransformationFilter()
        if data_set_external_id is not None:
            filter.data_set_ids = [ExternalId(external_id=data_set_external_id)]
        for transformations in self.client.tool.transformations.iterate(limit=None, filter=filter):
            yield from transformations

    def sensitive_strings(self, item: TransformationRequest) -> Iterable[str]:
        external_id = item.external_id
        if read_credentials := self._authentication_by_id_operation.get((external_id, "read")):
            yield read_credentials.client_secret
        if write_credentials := self._authentication_by_id_operation.get((external_id, "write")):
            yield write_credentials.client_secret


@final
class TransformationScheduleCRUD(
    ResourceCRUD[
        ExternalId,
        TransformationScheduleRequest,
        TransformationScheduleResponse,
    ]
):
    folder_name = "transformations"
    resource_cls = TransformationScheduleResponse
    resource_write_cls = TransformationScheduleRequest
    kind = "Schedule"
    yaml_cls = TransformationScheduleYAML
    dependencies = frozenset({TransformationCRUD})
    _doc_url = "Transformation-Schedules/operation/createTransformationSchedules"
    parent_resource = frozenset({TransformationCRUD})

    @property
    def display_name(self) -> str:
        return "transformation schedules"

    @classmethod
    def get_required_capability(
        cls, items: Sequence[TransformationScheduleRequest] | None, read_only: bool
    ) -> list[Capability]:
        # Access for transformations schedules is checked by the transformation that is deployed
        # first, so we don't need to check for any capabilities here.
        return []

    @classmethod
    def get_id(cls, item: TransformationScheduleResponse | TransformationScheduleRequest | dict) -> ExternalId:
        if isinstance(item, dict):
            return ExternalId(external_id=item["externalId"])
        if item.external_id is None:
            raise ToolkitRequiredValueError("TransformationSchedule must have external_id set.")
        return ExternalId(external_id=item.external_id)

    @classmethod
    def dump_id(cls, id: ExternalId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def as_str(cls, id: ExternalId) -> str:
        return sanitize_filename(id.external_id)

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "externalId" in item:
            yield TransformationCRUD, ExternalId(external_id=item["externalId"])

    def create(self, items: Sequence[TransformationScheduleRequest]) -> list[TransformationScheduleResponse]:
        try:
            return self.client.tool.transformations.schedules.create(list(items))
        except ToolkitAPIError as e:
            if not e.duplicated:
                raise
            existing = {external_id for dup in e.duplicated if (external_id := dup.get("externalId", None))}
            print(
                f"  [bold yellow]WARNING:[/] {len(e.duplicated)} transformation schedules already exist(s): {existing}"
            )
            new_items = [item for item in items if item.external_id not in existing]
            return self.client.tool.transformations.schedules.create(new_items)

    def retrieve(self, ids: Sequence[ExternalId]) -> list[TransformationScheduleResponse]:
        return self.client.tool.transformations.schedules.retrieve(list(ids), ignore_unknown_ids=True)

    def update(self, items: Sequence[TransformationScheduleRequest]) -> list[TransformationScheduleResponse]:
        return self.client.tool.transformations.schedules.update(list(items), mode="replace")

    def delete(self, ids: Sequence[ExternalId]) -> int:
        if not ids:
            return 0
        self.client.tool.transformations.schedules.delete(list(ids), ignore_unknown_ids=True)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[TransformationScheduleResponse]:
        if not parent_ids:
            for schedules in self.client.tool.transformations.schedules.iterate(limit=None):
                yield from schedules
        else:
            transformation_external_ids = [parent_id for parent_id in parent_ids if isinstance(parent_id, ExternalId)]
            if transformation_external_ids:
                yield from self.client.tool.transformations.schedules.retrieve(
                    transformation_external_ids, ignore_unknown_ids=True
                )


@final
class TransformationNotificationCRUD(
    ResourceCRUD[
        TransformationNotificationId,
        TransformationNotificationRequest,
        TransformationNotificationResponse,
    ]
):
    folder_name = "transformations"
    resource_cls = TransformationNotificationResponse
    resource_write_cls = TransformationNotificationRequest
    kind = "Notification"
    dependencies = frozenset({TransformationCRUD})
    _doc_url = "Transformation-Notifications/operation/createTransformationNotifications"
    parent_resource = frozenset({TransformationCRUD})
    yaml_cls = TransformationNotificationYAML

    support_update = False

    @property
    def display_name(self) -> str:
        return "transformation notifications"

    @classmethod
    def get_id(
        cls, item: TransformationNotificationResponse | TransformationNotificationRequest | dict
    ) -> TransformationNotificationId:
        if isinstance(item, dict):
            if missing := tuple(k for k in {"transformationExternalId", "destination"} if k not in item):
                # We need to raise a KeyError with all missing keys to get the correct error message.
                raise KeyError(*missing)
            return TransformationNotificationId(
                transformation_external_id=item["transformationExternalId"], destination=item["destination"]
            )

        return TransformationNotificationId(
            transformation_external_id=item.transformation_external_id or "<missing>",
            destination=item.destination,
        )

    @classmethod
    def dump_id(cls, id: TransformationNotificationId) -> dict[str, Any]:
        return id.dump()

    @classmethod
    def get_required_capability(
        cls, items: Sequence[TransformationNotificationRequest] | None, read_only: bool
    ) -> Capability | list[Capability]:
        # Access for transformation notification is checked by the transformation that is deployed
        # first, so we don't need to check for any capabilities here.
        return []

    def dump_resource(
        self, resource: TransformationNotificationResponse, local: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        dumped = resource.as_request_resource().dump()
        if local and "transformationExternalId" in local:
            dumped.pop("transformationId", None)
            dumped["transformationExternalId"] = local["transformationExternalId"]
        return dumped

    @classmethod
    def as_str(cls, id: TransformationNotificationId) -> str:
        return sanitize_filename(f"{id.transformation_external_id}_{id.destination}")

    def create(self, items: Sequence[TransformationNotificationRequest]) -> list[TransformationNotificationResponse]:
        return self.client.tool.transformations.notifications.create(list(items))

    def retrieve(self, ids: Sequence[TransformationNotificationId]) -> list[TransformationNotificationResponse]:
        unique_ids: set[ExternalId] = {ExternalId(external_id=id_.transformation_external_id) for id_ in ids}
        targets_ids = {(id_.transformation_external_id, id_.destination) for id_ in ids}
        return [
            notification
            for notification in self.iterate(parent_ids=list(unique_ids))
            if (notification.transformation_external_id, notification.destination) in targets_ids
        ]

    def delete(self, ids: Sequence[TransformationNotificationId]) -> int:
        # Note that it is theoretically possible that more items will be deleted than
        # input ids. This is because TransformationNotifications are identified by an internal id,
        # while the toolkit uses the transformationExternalId + destination as the id. Thus, there could
        # be multiple notifications for the same transformationExternalId + destination.
        if existing := self.retrieve(ids):
            self.client.tool.transformations.notifications.delete(InternalId.from_ids([item.id for item in existing]))
        return len(existing)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: Sequence[Hashable] | None = None,
    ) -> Iterable[TransformationNotificationResponse]:
        if parent_ids is None:
            for notifications in self.client.tool.transformations.notifications.iterate(limit=None):
                yield from notifications
        else:
            for parent_id in parent_ids:
                if not isinstance(parent_id, ExternalId):
                    continue
                filter = TransformationNotificationFilter(transformation_external_id=parent_id.external_id)
                for notifications in self.client.tool.transformations.notifications.iterate(limit=None, filter=filter):
                    for notification in notifications:
                        # This is not set by the API.
                        notification.transformation_external_id = parent_id.external_id
                        yield notification

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "transformationExternalId" in item:
            yield TransformationCRUD, ExternalId(external_id=item["transformationExternalId"])



from abc import ABC
from typing import Annotated, Any, Literal

from pydantic import Field, TypeAdapter, field_serializer
from pydantic_core.core_schema import FieldSerializationInfo

from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject

from ._references import ContainerReference, ViewReference


class PropertyTypeDefinition(BaseModelObject, ABC):
    type: str


class ListablePropertyTypeDefinition(PropertyTypeDefinition, ABC):
    list: bool | None = None
    max_list_size: int | None = None


class TextProperty(ListablePropertyTypeDefinition):
    type: Literal["text"] = "text"
    max_text_size: int | None = None
    collation: str | None = None


class Unit(BaseModelObject):
    external_id: str
    source_unit: str | None = None


class FloatProperty(ListablePropertyTypeDefinition, ABC):
    unit: Unit | None = None


class Float32Property(FloatProperty):
    type: Literal["float32"] = "float32"


class Float64Property(FloatProperty):
    type: Literal["float64"] = "float64"


class BooleanProperty(ListablePropertyTypeDefinition):
    type: Literal["boolean"] = "boolean"


class Int32Property(ListablePropertyTypeDefinition):
    type: Literal["int32"] = "int32"


class Int64Property(ListablePropertyTypeDefinition):
    type: Literal["int64"] = "int64"


class TimestampProperty(ListablePropertyTypeDefinition):
    type: Literal["timestamp"] = "timestamp"


class DateProperty(ListablePropertyTypeDefinition):
    type: Literal["date"] = "date"


class JSONProperty(ListablePropertyTypeDefinition):
    type: Literal["json"] = "json"


class TimeseriesCDFExternalIdReference(ListablePropertyTypeDefinition):
    type: Literal["timeseries"] = "timeseries"


class FileCDFExternalIdReference(ListablePropertyTypeDefinition):
    type: Literal["file"] = "file"


class SequenceCDFExternalIdReference(ListablePropertyTypeDefinition):
    type: Literal["sequence"] = "sequence"


class DirectNodeRelation(ListablePropertyTypeDefinition):
    type: Literal["direct"] = "direct"
    container: ContainerReference | None = None
    # This property is only available in the response object. It will be ignored in the request object.
    # In the request object, use ViewCoreProperty.source instead.
    source: ViewReference | None = Field(None, exclude=True)

    @field_serializer("container", mode="plain", when_used="unless-none")
    @classmethod
    def serialize_require(cls, container: ContainerReference, info: FieldSerializationInfo) -> dict[str, Any]:
        return {**container.model_dump(**vars(info)), "type": "container"}


class EnumValue(BaseModelObject):
    name: str | None = None
    description: str | None = None


class EnumProperty(PropertyTypeDefinition):
    type: Literal["enum"] = "enum"
    unknown_value: str | None = None
    values: dict[str, EnumValue]


DataType = Annotated[
    TextProperty
    | Float32Property
    | Float64Property
    | BooleanProperty
    | Int32Property
    | Int64Property
    | TimestampProperty
    | DateProperty
    | JSONProperty
    | TimeseriesCDFExternalIdReference
    | FileCDFExternalIdReference
    | SequenceCDFExternalIdReference
    | EnumProperty
    | DirectNodeRelation,
    Field(discriminator="type"),
]

DataTypeAdapter: TypeAdapter[DataType] = TypeAdapter(DataType)



from cognite_toolkit._cdf_tk.client._resource_base import BaseModelObject, RequestResource, ResponseResource

from .identifiers import ExternalId


class SimulatorRoutine(BaseModelObject):
    external_id: str
    model_external_id: str
    simulator_integration_external_id: str
    name: str
    description: str | None = None


class SimulatorRoutineRequest(RequestResource, SimulatorRoutine):
    """Request class for creating or updating a simulator routine."""

    def as_id(self) -> ExternalId:
        return ExternalId(external_id=self.external_id)


class SimulatorRoutineResponse(ResponseResource[SimulatorRoutineRequest], SimulatorRoutine):
    """Response class for a simulator routine."""

    id: int
    simulator_external_id: str
    description: str | None = None
    data_set_id: int
    created_time: int
    last_updated_time: int

    @classmethod
    def request_cls(cls) -> type[SimulatorRoutineRequest]:
        return SimulatorRoutineRequest



"""Group data classes for the Cognite API Groups endpoint.

Based on the API specification at:
https://api-docs.cognite.com/20230101/tag/Groups/operation/createGroups
"""

from .acls import (
    Acl,
    AgentsAcl,
    AnalyticsAcl,
    AnnotationsAcl,
    AppConfigAcl,
    AssetsAcl,
    AuditlogAcl,
    DataModelInstancesAcl,
    DataModelsAcl,
    DataSetsAcl,
    DiagramParsingAcl,
    DigitalTwinAcl,
    DocumentFeedbackAcl,
    DocumentPipelinesAcl,
    EntityMatchingAcl,
    EventsAcl,
    ExperimentsAcl,
    ExtractionConfigsAcl,
    ExtractionPipelinesAcl,
    ExtractionRunsAcl,
    FilePipelinesAcl,
    FilesAcl,
    FunctionsAcl,
    GeospatialAcl,
    GeospatialCrsAcl,
    GroupsAcl,
    HostedExtractorsAcl,
    LabelsAcl,
    LegacyGenericsAcl,
    LegacyModelHostingAcl,
    LocationFiltersAcl,
    MonitoringTasksAcl,
    NotificationsAcl,
    PipelinesAcl,
    PostgresGatewayAcl,
    ProjectsAcl,
    RawAcl,
    RelationshipsAcl,
    RoboticsAcl,
    SAPWritebackAcl,
    SAPWritebackRequestsAcl,
    ScheduledCalculationsAcl,
    SecurityCategoriesAcl,
    SeismicAcl,
    SequencesAcl,
    SessionsAcl,
    SimulatorsAcl,
    StreamRecordsAcl,
    StreamsAcl,
    TemplateGroupsAcl,
    TemplateInstancesAcl,
    ThreeDAcl,
    TimeSeriesAcl,
    TimeSeriesSubscriptionsAcl,
    TransformationsAcl,
    TypesAcl,
    UnknownAcl,
    UserProfilesAcl,
    VideoStreamingAcl,
    VisionModelAcl,
    WellsAcl,
    WorkflowOrchestrationAcl,
)
from .capability import (
    GroupCapability,
    ProjectUrlNames,
)
from .group import (
    Group,
    GroupAttributes,
    GroupRequest,
    GroupResponse,
    TokenAttributes,
)
from .scopes import (
    AllScope,
    AppConfigScope,
    AssetRootIDScope,
    CurrentUserScope,
    DataSetScope,
    ExperimentScope,
    ExtractionPipelineScope,
    IDScope,
    IDScopeLowerCase,
    InstancesScope,
    PartitionScope,
    PostgresGatewayUsersScope,
    Scope,
    ScopeDefinition,
    SpaceIDScope,
    TableScope,
)

__all__ = [
    "Acl",
    "AgentsAcl",
    "AllScope",
    "AnalyticsAcl",
    "AnnotationsAcl",
    "AppConfigAcl",
    "AppConfigScope",
    "AssetRootIDScope",
    "AssetsAcl",
    "AuditlogAcl",
    "CurrentUserScope",
    "DataModelInstancesAcl",
    "DataModelsAcl",
    "DataSetScope",
    "DataSetsAcl",
    "DiagramParsingAcl",
    "DigitalTwinAcl",
    "DocumentFeedbackAcl",
    "DocumentPipelinesAcl",
    "EntityMatchingAcl",
    "EventsAcl",
    "ExperimentScope",
    "ExperimentsAcl",
    "ExtractionConfigsAcl",
    "ExtractionPipelineScope",
    "ExtractionPipelinesAcl",
    "ExtractionRunsAcl",
    "FilePipelinesAcl",
    "FilesAcl",
    "FunctionsAcl",
    "GeospatialAcl",
    "GeospatialCrsAcl",
    "Group",
    "GroupAttributes",
    "GroupCapability",
    "GroupRequest",
    "GroupResponse",
    "GroupsAcl",
    "HostedExtractorsAcl",
    "IDScope",
    "IDScopeLowerCase",
    "InstancesScope",
    "LabelsAcl",
    "LegacyGenericsAcl",
    "LegacyModelHostingAcl",
    "LocationFiltersAcl",
    "MonitoringTasksAcl",
    "NotificationsAcl",
    "PartitionScope",
    "PipelinesAcl",
    "PostgresGatewayAcl",
    "PostgresGatewayUsersScope",
    "ProjectUrlNames",
    "ProjectsAcl",
    "RawAcl",
    "RelationshipsAcl",
    "RoboticsAcl",
    "SAPWritebackAcl",
    "SAPWritebackRequestsAcl",
    "ScheduledCalculationsAcl",
    "Scope",
    "ScopeDefinition",
    "SecurityCategoriesAcl",
    "SeismicAcl",
    "SequencesAcl",
    "SessionsAcl",
    "SimulatorsAcl",
    "SpaceIDScope",
    "StreamRecordsAcl",
    "StreamsAcl",
    "TableScope",
    "TemplateGroupsAcl",
    "TemplateInstancesAcl",
    "ThreeDAcl",
    "TimeSeriesAcl",
    "TimeSeriesSubscriptionsAcl",
    "TokenAttributes",
    "TransformationsAcl",
    "TypesAcl",
    "UnknownAcl",
    "UserProfilesAcl",
    "VideoStreamingAcl",
    "VisionModelAcl",
    "WellsAcl",
    "WorkflowOrchestrationAcl",
]



from collections.abc import Iterable, Iterator, Mapping, Sequence
from typing import ClassVar, Literal, cast

from cognite.client.data_classes.data_modeling import NodeId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.http_client import (
    HTTPClient,
    RequestMessage,
    ToolkitAPIError,
)
from cognite_toolkit._cdf_tk.client.http_client._item_classes import (
    ItemsFailedResponse,
    ItemsRequest,
    ItemsResultList,
    ItemsSuccessResponse,
)
from cognite_toolkit._cdf_tk.client.resource_classes.annotation import AnnotationResponse
from cognite_toolkit._cdf_tk.client.resource_classes.data_modeling import EdgeReference, InstanceRequest, NodeReference
from cognite_toolkit._cdf_tk.client.resource_classes.identifiers import InternalId
from cognite_toolkit._cdf_tk.client.resource_classes.legacy.pending_instances_ids import PendingInstanceId
from cognite_toolkit._cdf_tk.client.resource_classes.three_d import (
    AssetMappingClassicResponse,
    AssetMappingDMRequest,
    ThreeDModelClassicResponse,
)
from cognite_toolkit._cdf_tk.commands._migrate.data_classes import ThreeDMigrationRequest
from cognite_toolkit._cdf_tk.constants import MISSING_EXTERNAL_ID, MISSING_INSTANCE_SPACE
from cognite_toolkit._cdf_tk.exceptions import ToolkitNotImplementedError, ToolkitValueError
from cognite_toolkit._cdf_tk.storageio import (
    AnnotationIO,
    HierarchyIO,
    InstanceIO,
    T_Selector,
    UploadableStorageIO,
)
from cognite_toolkit._cdf_tk.storageio._base import Page, UploadItem
from cognite_toolkit._cdf_tk.storageio.selectors import (
    ThreeDModelFilteredSelector,
    ThreeDModelIdSelector,
    ThreeDSelector,
)
from cognite_toolkit._cdf_tk.tk_warnings import MediumSeverityWarning
from cognite_toolkit._cdf_tk.utils.collection import chunker_sequence
from cognite_toolkit._cdf_tk.utils.useful_types import (
    AssetCentricKindExtended,
    AssetCentricType,
    JsonVal,
)
from cognite_toolkit._cdf_tk.utils.useful_types2 import T_AssetCentricResource

from .data_classes import (
    AnnotationMapping,
    AssetCentricMapping,
    AssetCentricMappingList,
    MigrationMapping,
    MigrationMappingList,
)
from .data_model import INSTANCE_SOURCE_VIEW_ID
from .default_mappings import ASSET_ANNOTATIONS_ID, FILE_ANNOTATIONS_ID
from .selectors import AssetCentricMigrationSelector, MigrateDataSetSelector, MigrationCSVFileSelector


class AssetCentricMigrationIO(
    UploadableStorageIO[AssetCentricMigrationSelector, AssetCentricMapping[T_AssetCentricResource], InstanceRequest]
):
    KIND = "AssetCentricMigration"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson", ".yaml", ".yml"})
    CHUNK_SIZE = 1000
    UPLOAD_ENDPOINT = InstanceIO.UPLOAD_ENDPOINT

    PENDING_INSTANCE_ID_ENDPOINT_BY_KIND: ClassVar[Mapping[AssetCentricKindExtended, str]] = {
        "TimeSeries": "/timeseries/set-pending-instance-ids",
        "FileMetadata": "/files/set-pending-instance-ids",
    }

    def __init__(self, client: ToolkitClient, skip_linking: bool = True) -> None:
        super().__init__(client)
        self.hierarchy = HierarchyIO(client)
        self.skip_linking = skip_linking

    def as_id(self, item: AssetCentricMapping) -> str:
        return str(item.mapping.as_asset_centric_id())

    def stream_data(self, selector: AssetCentricMigrationSelector, limit: int | None = None) -> Iterator[Page]:
        if isinstance(selector, MigrationCSVFileSelector):
            iterator = self._stream_from_csv(selector, limit)
        elif isinstance(selector, MigrateDataSetSelector):
            iterator = self._stream_given_dataset(selector, limit)
        else:
            raise ToolkitNotImplementedError(f"Selector {type(selector)} is not supported for stream_data")
        yield from (Page(worker_id="main", items=items) for items in iterator)

    def _stream_from_csv(
        self, selector: MigrationCSVFileSelector, limit: int | None = None
    ) -> Iterator[Sequence[AssetCentricMapping[T_AssetCentricResource]]]:
        items = selector.items
        if limit is not None:
            items = MigrationMappingList(items[:limit])
        chunk: list[AssetCentricMapping[T_AssetCentricResource]] = []
        for current_batch in chunker_sequence(items, self.CHUNK_SIZE):
            resources = self.hierarchy.get_resource_io(selector.kind).retrieve(current_batch.get_ids())
            for mapping, resource in zip(current_batch, resources, strict=True):
                chunk.append(AssetCentricMapping(mapping=mapping, resource=resource))
            if chunk:
                yield chunk
                chunk = []

    def count(self, selector: AssetCentricMigrationSelector) -> int | None:
        if isinstance(selector, MigrationCSVFileSelector):
            return len(selector.items)
        elif isinstance(selector, MigrateDataSetSelector):
            return self.hierarchy.count(selector.as_asset_centric_selector())
        else:
            raise ToolkitNotImplementedError(f"Selector {type(selector)} is not supported for count")

    def _stream_given_dataset(
        self, selector: MigrateDataSetSelector, limit: int | None = None
    ) -> Iterator[Sequence[AssetCentricMapping[T_AssetCentricResource]]]:
        asset_centric_selector = selector.as_asset_centric_selector()
        for data_chunk in self.hierarchy.stream_data(asset_centric_selector, limit):
            mapping_list = AssetCentricMappingList[T_AssetCentricResource]([])
            for resource in data_chunk.items:
                # We got the resource from a dataset selector, so we know it is there
                data_set_id = cast(int, resource.data_set_id)
                space_source = self.client.migration.space_source.retrieve(data_set_id=data_set_id)
                instance_space = space_source.instance_space if space_source else None
                if instance_space is None:
                    instance_space = MISSING_INSTANCE_SPACE
                external_id = resource.external_id
                if external_id is None:
                    external_id = MISSING_EXTERNAL_ID.format(project=self.client.config.project, id=resource.id)
                mapping = MigrationMapping(
                    resource_type=self._kind_to_resource_type(selector.kind),
                    instance_id=NodeReference(
                        space=instance_space,
                        external_id=external_id,
                    ),
                    id=resource.id,
                    data_set_id=resource.data_set_id,
                    ingestion_view=selector.ingestion_mapping,
                    preferred_consumer_view=selector.preferred_consumer_view,
                )
                mapping_list.append(AssetCentricMapping(mapping=mapping, resource=resource))
            yield mapping_list

    @staticmethod
    def _kind_to_resource_type(kind: AssetCentricKindExtended) -> AssetCentricType:
        mapping: dict[AssetCentricKindExtended, AssetCentricType] = {
            "Assets": "asset",
            "Events": "event",
            "TimeSeries": "timeseries",
            "FileMetadata": "file",
        }
        try:
            return mapping[kind]
        except KeyError as e:
            raise ToolkitNotImplementedError(f"Kind '{kind}' is not supported") from e

    def data_to_json_chunk(
        self,
        data_chunk: Sequence[AssetCentricMapping[T_AssetCentricResource]],
        selector: AssetCentricMigrationSelector | None = None,
    ) -> list[dict[str, JsonVal]]:
        return [item.dump() for item in data_chunk]

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> InstanceRequest:
        raise NotImplementedError()

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[InstanceRequest]],
        http_client: HTTPClient,
        selector: AssetCentricMigrationSelector | None = None,
    ) -> ItemsResultList:
        """Upload items by first linking them using files/set-pending-instance-ids and then uploading the instances."""
        if self.skip_linking:
            return super().upload_items(data_chunk, http_client, None)
        elif selector is None:
            raise ToolkitNotImplementedError(f"Selector must be provided for uploading {self.KIND} items.")
        elif selector.kind not in self.PENDING_INSTANCE_ID_ENDPOINT_BY_KIND:
            return super().upload_items(data_chunk, http_client, None)

        pending_instance_id_endpoint = self.PENDING_INSTANCE_ID_ENDPOINT_BY_KIND[selector.kind]
        results = ItemsResultList()
        to_upload = self.link_asset_centric(data_chunk, http_client, pending_instance_id_endpoint)
        if to_upload:
            results.extend(super().upload_items(to_upload, http_client, None))
        return results

    @classmethod
    def link_asset_centric(
        cls,
        data_chunk: Sequence[UploadItem[InstanceRequest]],
        http_client: HTTPClient,
        pending_instance_id_endpoint: str,
    ) -> Sequence[UploadItem[InstanceRequest]]:
        """Links asset-centric resources to their (uncreated) instances using the pending-instance-ids endpoint."""
        config = http_client.config
        successful_linked: set[str] = set()
        for batch in chunker_sequence(data_chunk, cls.CHUNK_SIZE):
            batch_results = http_client.request_items_retries(
                message=ItemsRequest(
                    endpoint_url=config.create_api_url(pending_instance_id_endpoint),
                    method="POST",
                    api_version="alpha",
                    items=[
                        UploadItem(source_id=item.source_id, item=cls.as_pending_instance_id(item.item))
                        for item in batch
                    ],
                )
            )
            for res in batch_results:
                if isinstance(res, ItemsSuccessResponse):
                    successful_linked.update(res.ids)
        to_upload = [item for item in data_chunk if item.source_id in successful_linked]
        return to_upload

    @staticmethod
    def as_pending_instance_id(item: InstanceRequest) -> PendingInstanceId:
        """Convert an InstanceApply to a PendingInstanceId for linking."""
        source = next((source for source in item.sources or [] if source.source == INSTANCE_SOURCE_VIEW_ID), None)
        if source is None:
            raise ValueError(f"Cannot extract ID from item of type {type(item).__name__!r}")
        if source.properties is None:
            raise ValueError("Source properties cannot be None when linking asset-centric resources.")
        if not isinstance(source.properties["id"], int):
            raise ValueError(f"Unexpected ID type: {type(source.properties['id']).__name__!r}")
        id_ = source.properties["id"]
        return PendingInstanceId(
            pending_instance_id=NodeId(item.space, item.external_id),
            id=id_,
        )


class AnnotationMigrationIO(
    UploadableStorageIO[AssetCentricMigrationSelector, AssetCentricMapping[AnnotationResponse], InstanceRequest]
):
    """IO class for migrating Annotations.

    Args:
        client: The ToolkitClient to use for CDF interactions.
        instance_space: The instance space to use for the migrated annotations.
        default_asset_annotation_mapping: The default ingestion mapping to use for asset-linked annotations.
        default_file_annotation_mapping: The default ingestion mappingto use for file-linked annotations.

    """

    KIND = "AnnotationMigration"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".parquet", ".csv", ".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".parquet", ".csv", ".ndjson", ".yaml", ".yml"})
    CHUNK_SIZE = 1000
    UPLOAD_ENDPOINT = InstanceIO.UPLOAD_ENDPOINT

    SUPPORTED_ANNOTATION_TYPES = frozenset({"diagrams.AssetLink", "diagrams.FileLink"})

    def __init__(
        self,
        client: ToolkitClient,
        instance_space: str | None = None,
        default_asset_annotation_mapping: str | None = None,
        default_file_annotation_mapping: str | None = None,
    ) -> None:
        super().__init__(client)
        self.annotation_io = AnnotationIO(client)
        self.instance_space = instance_space
        self.default_asset_annotation_mapping = default_asset_annotation_mapping or ASSET_ANNOTATIONS_ID
        self.default_file_annotation_mapping = default_file_annotation_mapping or FILE_ANNOTATIONS_ID

    def as_id(self, item: AssetCentricMapping[AnnotationResponse]) -> str:
        return f"Annotation_{item.mapping.id}"

    def count(self, selector: AssetCentricMigrationSelector) -> int | None:
        if isinstance(selector, MigrationCSVFileSelector):
            return len(selector.items)
        else:
            # There is no efficient way to count annotations in CDF.
            return None

    def stream_data(self, selector: AssetCentricMigrationSelector, limit: int | None = None) -> Iterable[Page]:
        if isinstance(selector, MigrateDataSetSelector):
            iterator = self._stream_from_dataset(selector, limit)
        elif isinstance(selector, MigrationCSVFileSelector):
            iterator = self._stream_from_csv(selector, limit)
        else:
            raise ToolkitNotImplementedError(f"Selector {type(selector)} is not supported for stream_data")
        yield from (Page(worker_id="main", items=items) for items in iterator)

    def _stream_from_dataset(
        self, selector: MigrateDataSetSelector, limit: int | None = None
    ) -> Iterator[Sequence[AssetCentricMapping[AnnotationResponse]]]:
        if self.instance_space is None:
            raise ToolkitValueError("Instance space must be provided for dataset-based annotation migration.")
        asset_centric_selector = selector.as_asset_centric_selector()
        for data_chunk in self.annotation_io.stream_data(asset_centric_selector, limit):
            mapping_list = AssetCentricMappingList[AnnotationResponse]([])
            for resource in data_chunk.items:
                if resource.annotation_type not in self.SUPPORTED_ANNOTATION_TYPES:
                    # This should not happen, as the annotation_io should already filter these out.
                    # This is just in case.
                    continue
                mapping = AnnotationMapping(
                    instance_id=EdgeReference(space=self.instance_space, external_id=f"annotation_{resource.id!r}"),
                    id=resource.id,
                    ingestion_view=self._get_mapping(selector.ingestion_mapping, resource),
                    preferred_consumer_view=selector.preferred_consumer_view,
                    annotation_type=resource.annotation_type,  # type: ignore[arg-type]
                )
                mapping_list.append(AssetCentricMapping(mapping=mapping, resource=resource))
            yield mapping_list

    def _stream_from_csv(
        self, selector: MigrationCSVFileSelector, limit: int | None = None
    ) -> Iterator[Sequence[AssetCentricMapping[AnnotationResponse]]]:
        items = selector.items
        if limit is not None:
            items = MigrationMappingList(items[:limit])
        chunk: list[AssetCentricMapping[AnnotationResponse]] = []
        for current_batch in chunker_sequence(items, self.CHUNK_SIZE):
            resources = self.client.tool.annotations.retrieve([InternalId(id=id_) for id_ in current_batch.get_ids()])
            resources_by_id = {resource.id: resource for resource in resources}
            not_found = 0
            incorrect_type_count = 0
            for mapping in current_batch:
                resource = resources_by_id.get(mapping.id)
                if resource is None:
                    not_found += 1
                    continue
                if resource.annotation_type not in self.SUPPORTED_ANNOTATION_TYPES:
                    incorrect_type_count += 1
                    continue
                mapping.ingestion_view = self._get_mapping(mapping.ingestion_view, resource)
                chunk.append(AssetCentricMapping(mapping=mapping, resource=resource))
            if chunk:
                yield chunk
                chunk = []
            if not_found:
                MediumSeverityWarning(
                    f"Could not find {not_found} annotations referenced in the CSV file. They will be skipped during migration."
                ).print_warning(include_timestamp=True, console=self.client.console)
            if incorrect_type_count:
                MediumSeverityWarning(
                    f"Found {incorrect_type_count} annotations with unsupported types. Only 'diagrams.AssetLink' and "
                    "'diagrams.FileLink' are supported. These annotations will be skipped during migration."
                ).print_warning(include_timestamp=True, console=self.client.console)

    def _get_mapping(self, current_mapping: str | None, resource: AnnotationResponse) -> str:
        try:
            return (
                current_mapping
                or {
                    "diagrams.AssetLink": self.default_asset_annotation_mapping,
                    "diagrams.FileLink": self.default_file_annotation_mapping,
                }[resource.annotation_type]
            )
        except KeyError as e:
            raise ToolkitValueError(
                f"Could not determine default ingestion view for annotation type '{resource.annotation_type}'. "
                "Please specify the ingestion view explicitly in the CSV file."
            ) from e

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> InstanceRequest:
        raise NotImplementedError("Deserializing Annotation Migrations from JSON is not supported.")

    def data_to_json_chunk(
        self,
        data_chunk: Sequence[AssetCentricMapping[AnnotationResponse]],
        selector: AssetCentricMigrationSelector | None = None,
    ) -> list[dict[str, JsonVal]]:
        raise NotImplementedError("Serializing Annotation Migrations to JSON is not supported.")


class ThreeDMigrationIO(UploadableStorageIO[ThreeDSelector, ThreeDModelClassicResponse, ThreeDMigrationRequest]):
    """IO class for downloading and migrating 3D models.

    Args:
        client: The ToolkitClient to use for CDF interactions.
        data_model_type: The type of 3D data model to download. Either "classic" or "DM".

    """

    KIND = "3DMigration"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    DOWNLOAD_LIMIT = 1000
    CHUNK_SIZE = 1
    UPLOAD_ENDPOINT = "/3d/migrate/models"
    REVISION_ENDPOINT = "/3d/migrate/revisions"

    def __init__(self, client: ToolkitClient, data_model_type: Literal["classic", "data modeling"] = "classic") -> None:
        super().__init__(client)
        self.data_model_type = data_model_type

    def as_id(self, item: ThreeDModelClassicResponse) -> str:
        return item.name

    def _is_selected(self, item: ThreeDModelClassicResponse, included_models: set[int] | None) -> bool:
        return self._is_correct_type(item) and (included_models is None or item.id in included_models)

    def _is_correct_type(self, item: ThreeDModelClassicResponse) -> bool:
        if self.data_model_type == "classic":
            return item.space is None
        else:
            return item.space is not None

    def stream_data(
        self, selector: ThreeDSelector, limit: int | None = None
    ) -> Iterable[Page[ThreeDModelClassicResponse]]:
        published: bool | None = None
        if isinstance(selector, ThreeDModelFilteredSelector):
            published = selector.published
        included_models: set[int] | None = None
        if isinstance(selector, ThreeDModelIdSelector):
            included_models = set(selector.ids)
        cursor: str | None = None
        total = 0
        while True:
            request_limit = min(self.DOWNLOAD_LIMIT, limit - total) if limit is not None else self.DOWNLOAD_LIMIT
            response = self.client.tool.three_d.models_classic.paginate(
                published=published, include_revision_info=True, limit=request_limit, cursor=cursor
            )
            items = [item for item in response.items if self._is_selected(item, included_models)]
            total += len(items)
            if items:
                yield Page(worker_id="main", items=items, next_cursor=response.next_cursor)
            if response.next_cursor is None:
                break
            cursor = response.next_cursor

    def count(self, selector: ThreeDSelector) -> int | None:
        # There is no efficient way to count 3D models in CDF.
        return None

    def data_to_json_chunk(
        self, data_chunk: Sequence[ThreeDModelClassicResponse], selector: ThreeDSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        raise NotImplementedError("Deserializing Annotation Migrations from JSON is not supported.")

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> ThreeDMigrationRequest:
        raise NotImplementedError("Deserializing ThreeD Migrations from JSON is not supported.")

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[ThreeDMigrationRequest]],
        http_client: HTTPClient,
        selector: ThreeDSelector | None = None,
    ) -> ItemsResultList:
        """Migrate 3D models by uploading them to the migrate/models endpoint."""
        if len(data_chunk) > self.CHUNK_SIZE:
            raise RuntimeError(f"Uploading more than {self.CHUNK_SIZE} 3D models at a time is not supported.")

        results = ItemsResultList()
        responses = http_client.request_items_retries(
            message=ItemsRequest(
                endpoint_url=self.client.config.create_api_url(self.UPLOAD_ENDPOINT),
                method="POST",
                items=data_chunk,
            )
        )
        if (
            failed_response := next((res for res in responses if isinstance(res, ItemsFailedResponse)), None)
        ) and failed_response.status_code == 400:
            raise ToolkitAPIError("3D model migration failed. You need to enable the 3D migration alpha feature flag.")

        results.extend(responses)
        success_ids = {id for res in responses if isinstance(res, ItemsSuccessResponse) for id in res.ids}
        for data in data_chunk:
            if data.source_id not in success_ids:
                continue
            revision = http_client.request_single_retries(
                message=RequestMessage(
                    endpoint_url=self.client.config.create_api_url(self.REVISION_ENDPOINT),
                    method="POST",
                    body_content={"items": [data.item.revision.dump(camel_case=True)]},
                )
            )
            results.append(revision.as_item_response(data.source_id))
        return results


class ThreeDAssetMappingMigrationIO(
    UploadableStorageIO[ThreeDSelector, AssetMappingClassicResponse, AssetMappingDMRequest]
):
    KIND = "3DMigrationAssetMapping"
    SUPPORTED_DOWNLOAD_FORMATS = frozenset({".ndjson"})
    SUPPORTED_COMPRESSIONS = frozenset({".gz"})
    SUPPORTED_READ_FORMATS = frozenset({".ndjson"})
    DOWNLOAD_LIMIT = 1000
    CHUNK_SIZE = 100
    UPLOAD_ENDPOINT = "/3d/models/{modelId}/revisions/{revisionId}/mappings"

    def __init__(self, client: ToolkitClient, object_3D_space: str, cad_node_space: str) -> None:
        super().__init__(client)
        self.object_3D_space = object_3D_space
        self.cad_node_space = cad_node_space
        # We can only migrate asset mappings for 3D models that are already migrated to data modeling.
        self._3D_io = ThreeDMigrationIO(client, data_model_type="data modeling")

    def as_id(self, item: AssetMappingClassicResponse) -> str:
        return f"AssetMapping_{item.model_id!s}_{item.revision_id!s}_{item.asset_id!s}"

    def stream_data(
        self, selector: ThreeDSelector, limit: int | None = None
    ) -> Iterable[Page[AssetMappingClassicResponse]]:
        total = 0
        for three_d_page in self._3D_io.stream_data(selector, None):
            for model in three_d_page.items:
                if model.last_revision_info is None or model.last_revision_info.revision_id is None:
                    # No revisions, so no asset mappings to
                    continue
                cursor: str | None = None
                while True:
                    request_limit = (
                        min(self.DOWNLOAD_LIMIT, limit - total) if limit is not None else self.DOWNLOAD_LIMIT
                    )
                    if limit is not None and total >= limit:
                        return
                    response = self.client.tool.three_d.asset_mappings_classic.paginate(
                        model_id=model.id,
                        revision_id=model.last_revision_info.revision_id,
                        cursor=cursor,
                        limit=request_limit,
                    )
                    items = response.items
                    total += len(items)
                    if items:
                        yield Page(worker_id="main", items=items, next_cursor=response.next_cursor)
                    if response.next_cursor is None:
                        break
                    cursor = response.next_cursor

    def count(self, selector: ThreeDSelector) -> int | None:
        # There is no efficient way to count 3D asset mappings in CDF.
        return None

    def upload_items(
        self,
        data_chunk: Sequence[UploadItem[AssetMappingDMRequest]],
        http_client: HTTPClient,
        selector: T_Selector | None = None,
    ) -> ItemsResultList:
        """Migrate 3D asset mappings by uploading them to the migrate/asset-mappings endpoint."""
        if not data_chunk:
            return ItemsResultList()
        # Assume all items in the chunk belong to the same model and revision, they should
        # if the .stream_data method is used for downloading.
        first = data_chunk[0]
        model_id = first.item.model_id
        revision_id = first.item.revision_id
        endpoint = self.UPLOAD_ENDPOINT.format(modelId=model_id, revisionId=revision_id)
        return http_client.request_items_retries(
            ItemsRequest(
                endpoint_url=self.client.config.create_api_url(endpoint),
                method="POST",
                items=data_chunk,
                extra_body_fields={
                    "dmsContextualizationConfig": {
                        "object3DSpace": self.object_3D_space,
                        "cadNodeSpace": self.cad_node_space,
                    }
                },
            )
        )

    def json_to_resource(self, item_json: dict[str, JsonVal]) -> AssetMappingDMRequest:
        raise NotImplementedError("Deserializing 3D Asset Mappings from JSON is not supported.")

    def data_to_json_chunk(
        self, data_chunk: Sequence[AssetMappingClassicResponse], selector: ThreeDSelector | None = None
    ) -> list[dict[str, JsonVal]]:
        raise NotImplementedError("Serializing 3D Asset Mappings to JSON is not supported.")
